import pytest

from sc3nb.helpers import ampdb, clip, cpsmidi, dbamp, linlin, midicps


def test_linlin():
    assert -50 == linlin(-1, 0, 2, 0, 100)
    assert 0 == linlin(0, 0, 2, 0, 100)
    assert 50 == linlin(1, 0, 2, 0, 100)
    assert 100 == linlin(2, 0, 2, 0, 100)
    assert 200 == linlin(4, 0, 2, 0, 100)
    assert 100 == linlin(3, 5, 105, 100, 1000, "min")


def test_clip():
    for x, y in zip([2, 3, 4, 5, 6], [3, 3, 4, 5, 5]):
        assert y == clip(x, 3, 5)


def test_midi_cps():
    for x in range(128):
        assert x == cpsmidi(midicps(x))


def test_db_amp():
    for x in range(128):
        assert x == pytest.approx(ampdb(dbamp(x)))



"""Classes for managing ID allocations."""

from abc import ABC, abstractmethod
from typing import Sequence


class Allocator(ABC):
    @abstractmethod
    def allocate(self, num: int = 1) -> Sequence[int]:
        raise NotImplementedError

    @abstractmethod
    def free(self, ids: Sequence[int]) -> None:
        raise NotImplementedError


class NodeAllocator(Allocator):
    """Allows allocating ids for Nodes."""

    def __init__(self, client_id: int) -> None:
        self.client_id = client_id
        self._num_node_ids = 0

    def allocate(self, num: int = 1) -> Sequence[int]:
        self._num_node_ids += 1
        if self._num_node_ids >= 2 ** 31:
            self._num_node_ids = 0
        return [self._num_node_ids + 10000 * (self.client_id + 1)]

    def free(self, ids: Sequence[int]) -> None:
        pass


class BlockAllocator(Allocator):
    """Allows allocating blocks of ids / indexes"""

    def __init__(self, num_ids: int, offset: int) -> None:
        self._offset = offset
        self._free_ids = [i + offset for i in range(num_ids)]

    def allocate(self, num: int = 1) -> Sequence[int]:
        """Allocate the next free ids

        Returns
        -------
        int
            free ids

        Raises
        ------
        RuntimeError
            When out of free ids or not enough ids are in order.
        """
        num_collected_ids = 1
        first_idx = 0
        idx = 0
        while num_collected_ids != num:
            if len(self._free_ids[first_idx:]) < num:
                raise RuntimeError(f"Cannot allocate {num} ids.")
            num_collected_ids = 1
            for idx in range(1, len(self._free_ids[first_idx:])):
                prev_id = self._free_ids[first_idx + idx - 1]
                next_id = self._free_ids[first_idx + idx]
                if abs(prev_id - next_id) > 1:
                    # difference between ids is too large
                    first_idx += idx
                    break
                num_collected_ids += 1
                if num_collected_ids == num:
                    break
        ids = self._free_ids[first_idx : first_idx + idx + 1]
        del self._free_ids[first_idx : first_idx + idx + 1]
        return ids

    def free(self, ids: Sequence[int]) -> None:
        """Mark ids as free again.

        Parameters
        ----------
        ids : sequence of int
            ids that are not used anymore.
        """
        for free_id in ids:
            self._free_ids.insert(free_id - self._offset, free_id)



import aifc
import tempfile
import time
from pathlib import Path

from sc3nb.osc.osc_communication import OSCMessage
from sc3nb.sc_objects.score import Score
from sc3nb.sc_objects.server import ServerOptions
from sc3nb.sc_objects.synthdef import SynthDef
from tests.conftest import SCBaseTest


class ScoreTest(SCBaseTest):

    __test__ = True
    start_sclang = True

    def setUp(self) -> None:
        self.assertIsNotNone(ScoreTest.sc.lang)
        synthdef = SynthDef(
            "test",
            r"""{ |out, freq = 440|
            OffsetOut.ar(out,
                SinOsc.ar(freq, 0, 0.2) * Line.kr(1, 0, 0.5, doneAction: Done.freeSelf)
            )
        }""",
        )
        with ScoreTest.sc.server.bundler(send_on_exit=False) as bundler:
            synthdef.add()
            bundler.add(0.0, "/s_new", ["test", 1003, 0, 0, "freq", 440])
            bundler.add(0.1, "/s_new", ["test", 1000, 0, 0, "freq", 440])
            bundler.add(0.1, "/s_new", ["test", 1004, 0, 0, "freq", 440])
            bundler.add(0.2, "/s_new", ["test", 1001, 0, 0, "freq", 660])
            bundler.add(0.3, "/s_new", ["test", 1002, 0, 0, "freq", 220])
            bundler.add(1, "/c_set", [0, 0])
        self.messages = bundler.messages()

    def test_score(self):
        sc3nb_osc = "test.osc"
        sc3nb_snd = "test.aiff"

        sclang_osc = "sclang.osc"
        sclang_snd = "sclang.aiff"

        with tempfile.TemporaryDirectory(dir=".") as tmp_path:
            tmp_path = ("." / Path(tmp_path)).resolve()
            cp = Score.record_nrt(
                self.messages,
                str(tmp_path / sc3nb_osc),
                str(tmp_path / sc3nb_snd),
                options=ServerOptions(num_output_buses=2),
                header_format="AIFF",
            )
            print(cp)
            (_, port), _ = ScoreTest.sc.server.connection_info()
            ScoreTest.sc.lang.cmd(
                fr"""
                var g;
                TempoClock.default.tempo = 1;
                d = SynthDef("test", {{ |out, freq = 440| OffsetOut.ar(out, SinOsc.ar(freq, 0, 0.2) * Line.kr(1, 0, 0.5, doneAction: Done.freeSelf))}});
                g = [
                    [0.0, [\d_recv, d.asBytes]],
                    [0.0, [\s_new, \test, 1003, 0, 0, \freq, 440]],
                    [0.1, [\s_new, \test, 1000, 0, 0, \freq, 440]],
                    [0.1, [\s_new, \test, 1004, 0, 0, \freq, 440]],
                    [0.2, [\s_new, \test, 1001, 0, 0, \freq, 660]],
                    [0.3, [\s_new, \test, 1002, 0, 0, \freq, 220]],
                    [1, [\c_set, 0, 0]] // finish
                    ];
                Score.recordNRT(
                    list: g,
                    oscFilePath: "{(tmp_path / sclang_osc).as_posix()}",
                    outputFilePath: "{(tmp_path / sclang_snd).as_posix()}",
                    headerFormat: "AIFF",
                    action: {{ NetAddr("127.0.0.1", {port}).sendMsg('/return', "nrt-done") }}
                );
                """
            )
            self.assertEqual(
                "nrt-done", ScoreTest.sc.server.msg_queues["/return"].get(timeout=4)
            )
            with aifc.open((tmp_path / sclang_snd).as_posix(), "rb") as sclang_wav:
                with aifc.open((tmp_path / sc3nb_snd).as_posix(), "rb") as sc3nb_wav:
                    self.assertEqual(sclang_wav.getparams(), sc3nb_wav.getparams())

            with self.assertWarnsRegex(
                UserWarning,
                "This method currently destroys the time tag if they happend in the past.",
            ):
                messages_sclang = Score.load_file(tmp_path / sclang_osc)
                messages_sc3nb = Score.load_file(tmp_path / sc3nb_osc)

            messages_sclang = next(iter(messages_sclang.values()))
            messages_sc3nb = next(iter(messages_sc3nb.values()))

            gmsg = OSCMessage("/g_new", [1, 0, 0])
            for msg in messages_sclang[:2]:
                self.assertTrue(
                    msg.address == gmsg.address and msg.parameters == gmsg.parameters
                )

            for n, msg in enumerate(messages_sclang[2:]):
                self.assertTrue(
                    msg.address == messages_sc3nb[n].address
                    and msg.parameters == messages_sc3nb[n].parameters
                )



import time
from queue import Empty

import pytest

from sc3nb.sc_objects.node import Group, Synth
from tests.conftest import SCBaseTest


class NodeTest(SCBaseTest):

    __test__ = True

    def setUp(self) -> None:
        with self.assertRaises(RuntimeError):
            NodeTest.sc.lang

    def test_wait(self):
        duration = 0.15
        tol = 0.05

        def check(synth):
            self.assertEqual(synth.is_playing, None)
            self.assertEqual(synth.started, True)
            self.assertEqual(synth.freed, False)
            self.assertEqual(synth.group, self.sc.server.default_group.nodeid)
            t_wait_for_notification = time.time()
            while not synth.is_playing:
                if time.time() - t_wait_for_notification > 0.15:
                    self.fail("Waiting for /n_go notification took too long.")
            self.assertEqual(synth.is_playing, True)
            self.assertEqual(synth.started, True)
            self.assertEqual(synth.freed, False)
            synth.wait(timeout=1)
            time_played = time.time() - t0
            self.assertEqual(synth.is_playing, False)
            self.assertEqual(synth.started, False)
            self.assertEqual(synth.freed, True)
            self.assertEqual(synth.group, None)
            self.assertLessEqual(time_played, duration + tol)
            self.assertGreaterEqual(time_played, duration - tol)

        t0 = time.time()
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's1' is unknown", msg="SynthDesc seems to be known"
        ):
            s1_synth = Synth("s1", controls={"dur": duration, "amp": 0.0})
        check(s1_synth)
        t0 = time.time()
        s1_synth.new()
        check(s1_synth)

    def test_fast_wait(self):
        duration = 0.15
        tol = 0.05

        def check(synth):
            self.assertEqual(synth.is_playing, False)
            self.assertEqual(synth.started, False)
            self.assertEqual(synth.freed, True)
            time_played = time.time() - t0
            self.assertLessEqual(time_played, duration + tol)
            self.assertGreaterEqual(time_played, duration - tol)

        t0 = time.time()
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's1' is unknown", msg="SynthDesc seems to be known"
        ):
            s1_synth = Synth("s1", controls={"dur": duration, "amp": 0.0})
        s1_synth.wait(timeout=1)
        check(s1_synth)

        t0 = time.time()
        s1_synth.new()
        s1_synth.wait(timeout=1)
        check(s1_synth)

    def test_too_many_arguments(self):
        with self.assertRaises(TypeError):
            Synth("s2", {"amp": 0.0}, "this is too much!")
        with self.assertRaises(TypeError):
            Group(101, "this is too much!")

    def test_reuse_nodeid(self):
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's2' is unknown", msg="SynthDesc seems to be known"
        ):
            synth1 = Synth("s2", {"amp": 0.0})
            nodeid = synth1.nodeid
            synth1.free()
            synth1.wait(timeout=1)
            group = Group(nodeid=nodeid)
            with self.assertRaisesRegex(RuntimeError, "Tried to get "):
                Synth("s2", nodeid=nodeid)
            group.free()
            group.wait(timeout=1)
            synth2 = Synth("s2", nodeid=nodeid)
            synth2.free()
            synth2.wait(timeout=1)

    @pytest.mark.allowloggingwarn
    def test_duplicate(self):
        self.assertNotIn("/s_new", self.sc.server.fails)
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's2' is unknown", msg="SynthDesc seems to be known"
        ):
            synth1 = Synth("s2", controls={"amp": 0.0})
            wait_t0 = time.time()
            synth1.new(controls={"amp": 0.0})
            while not "/s_new" in self.sc.server.fails:
                self.assertLessEqual(time.time() - wait_t0, 0.5)
            self.assertEqual(self.sc.server.fails["/s_new"].get(), "duplicate node ID")
            synth1.free()
            synth1.wait(timeout=1)
            synth1.new({"amp": 0.0})
            synth1.free()
            synth1.wait(timeout=1)
            with self.assertRaises(Empty):
                self.sc.server.fails["/s_new"].get(timeout=0.5)



import time

from sc3nb import Synth
from sc3nb.osc.osc_communication import Bundler
from tests.conftest import SCBaseTest


class BundlerTest(SCBaseTest):

    __test__ = True

    def test_context_manager(self):
        time_passed = 3.0
        with Bundler(send_on_exit=False) as bundler:
            BundlerTest.sc.server.msg("/status", bundle=True)
            BundlerTest.sc.server.msg("/status", bundle=False)
            bundler.wait(time_passed)
            BundlerTest.sc.server.msg("/status", bundle=True)
        self.assertEqual(len(bundler.contents), 2)
        self.assertEqual(bundler.passed_time, time_passed)

    def test_bundler_styles(self):
        BundlerTest.sc.server.latency = 0.1
        time_between = 0.2
        with BundlerTest.sc.server.bundler(
            send_on_exit=False
        ) as server_bundler_auto_bundling:
            BundlerTest.sc.server.msg("/status", bundle=True)
            server_bundler_auto_bundling.wait(time_between)
            BundlerTest.sc.server.msg("/status", bundle=True)

        with BundlerTest.sc.server.bundler(send_on_exit=False) as server_bundler_add:
            server_bundler_add.add(0.0, "/status")
            server_bundler_add.add(time_between, "/status")

        with Bundler(timetag=self.sc.server.latency, send_on_exit=False) as bundler_add:
            bundler_add.add(0.0, "/status")
            bundler_add.add(time_between, "/status")

        server_auto_bundle = server_bundler_auto_bundling.to_raw_osc(start_time=0)
        server_bundle = server_bundler_add.to_raw_osc(start_time=0)
        bundle = bundler_add.to_raw_osc(start_time=0)

        self.assertEqual(server_bundle, bundle)
        self.assertEqual(server_auto_bundle, bundle)

    def test_bundler_messages(self):
        start = 2
        stop = 7
        dur = 0.25
        num_abs_time = 5

        t0 = time.time()
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's.' is unknown", msg="SynthDesc seems to be known"
        ):
            with Bundler(send_on_exit=False) as bundler:
                for offset in range(1, 1 + num_abs_time):
                    with Bundler(t0 + offset):
                        Synth("s1", dict(freq=1000, dur=dur / 3))
                for n in range(start, stop):
                    dur = 0.25
                    Synth("s1", dict(freq=150 * n, dur=dur))
                    temp = Synth("s2", dict(freq=100 * n))
                    bundler.wait(dur * n / 2)
                    temp.release(dur)

        msgs = bundler.messages()
        # check if all messages are in the values
        self.assertEqual(len(msgs.values()), stop - start + 1 + num_abs_time)
        # check if all messages are in the right timetag entry
        self.assertEqual(
            list(map(len, list(msgs.values()))),
            num_abs_time * [1] + [2] + (stop - start - 1) * [3] + [1],
        )
        # check times
        expected_times = [1.0, 2.0, 3.0, 4.0, 5.0, 0.0, 0.25, 0.625, 1.125, 1.75, 2.5]
        self.assertEqual(
            [x - t0 if x > t0 else x for x in list(msgs.keys())], expected_times
        )  # need to add t0 here
        delay = 1
        msgs = bundler.messages(0, delay)
        self.assertEqual(
            [x - t0 if x > t0 else x for x in list(msgs.keys())],
            [tt + delay for tt in expected_times],
        )
        # check specific start
        msgs = bundler.messages(t0)
        self.assertEqual(
            list(msgs.keys()),
            [tt + t0 for tt in expected_times],
        )
        # check specific start + delay
        msgs = bundler.messages(t0, -t0)
        self.assertEqual(
            list(msgs.keys()),
            expected_times,
        )



"""Implements Node and subclasses Synth and Group."""

import logging
import warnings
from abc import ABC, abstractmethod
from enum import Enum, unique
from functools import reduce
from operator import iconcat
from threading import Event, RLock
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import sc3nb
from sc3nb.osc.osc_communication import OSCCommunicationError, OSCMessage
from sc3nb.sc_objects.synthdef import SynthDef

if TYPE_CHECKING:
    from sc3nb.sc_objects.bus import Bus
    from sc3nb.sc_objects.server import SCServer
    from sc3nb.sclang import SynthArgument

_LOGGER = logging.getLogger(__name__)


@unique
class GroupReply(str, Enum):
    """Replies of Group Commands"""

    QUERY_TREE_REPLY = "/g_queryTree.reply"


@unique
class GroupCommand(str, Enum):
    """OSC Commands for Groups"""

    QUERY_TREE = "/g_queryTree"
    DUMP_TREE = "/g_dumpTree"
    DEEP_FREE = "/g_deepFree"
    FREE_ALL = "/g_freeAll"
    TAIL = "/g_tail"
    HEAD = "/g_head"
    G_NEW = "/g_new"
    P_NEW = "/p_new"


@unique
class SynthCommand(str, Enum):
    """OSC Commands for Synths"""

    NEW = "/s_new"
    S_GET = "/s_get"
    S_GETN = "/s_getn"


@unique
class NodeReply(str, Enum):
    """Replies of Node Commands"""

    INFO = "/n_info"


@unique
class NodeCommand(str, Enum):
    """OSC Commands for Nodes"""

    ORDER = "/n_order"
    TRACE = "/n_trace"
    QUERY = "/n_query"
    MAP = "/n_map"
    MAPN = "/n_mapn"
    MAPA = "/n_mapa"
    MAPAN = "/n_mapan"
    FILL = "/n_fill"
    SET = "/n_set"
    SETN = "/n_setn"
    RUN = "/n_run"
    FREE = "/n_free"


@unique
class AddAction(Enum):
    """AddAction of SuperCollider nodes.

    This Enum contains the codes for the different ways to add a node.
    """

    TO_HEAD = 0  # (the default) add at the head of the group specified by target
    TO_TAIL = 1  # add at the tail of the group specified by target
    BEFORE = 2  # add immediately before target in its server's node order
    AFTER = 3  # add immediately after target in its server's node order
    REPLACE = 4  # replace target and take its place in its server's node order
    # Note: A Synth is not a valid target for \addToHead and \addToTail.


class SynthInfo(NamedTuple):
    """Information about the Synth from /n_info"""

    nodeid: int
    group: int
    prev_nodeid: int
    next_nodeid: int


class GroupInfo(NamedTuple):
    """Information about the Group from /n_info"""

    nodeid: int
    group: int
    prev_nodeid: int
    next_nodeid: int
    head: int
    tail: int


class Node(ABC):
    """Representation of a Node on SuperCollider."""

    def __new__(
        cls,
        *args: Any,
        nodeid: Optional[int] = None,
        server: Optional["SCServer"] = None,
        **kwargs: Any,
    ) -> "Node":
        if nodeid is not None:
            if server is None:
                server = sc3nb.SC.get_default().server
            try:
                node = server.nodes[nodeid]
                if node is not None:
                    if node.freed and not node.is_playing:
                        _LOGGER.debug(
                            "Removing %s(%s) from %s",
                            type(node).__name__,
                            node.nodeid,
                            repr(server),
                        )
                        del server.nodes[nodeid]
                    # check here if nodes types are compatible
                    elif isinstance(node, cls):
                        _LOGGER.debug(
                            "Returned %s(%s) from %s",
                            type(node).__name__,
                            node.nodeid,
                            repr(server),
                        )
                        return node
                    else:
                        raise RuntimeError(
                            f"Tried to get {node} from {server}"
                            f" as {cls.__name__} but type is {type(node).__name__}"
                        )
            except KeyError:
                pass
        _LOGGER.debug("%s(%s) not in Server (%s)", cls.__name__, nodeid, server)
        return super().__new__(cls)

    @abstractmethod
    def __init__(
        self,
        *,
        nodeid: Optional[int] = None,
        add_action: Optional[Union[AddAction, int]] = None,
        target: Optional[Union["Node", int]] = None,
        server: Optional["SCServer"] = None,
    ) -> None:
        """Create a new Node

        Parameters
        ----------
        nodeid : int or None
            This Nodes node id or None
        add_action : AddAction or corresponding int, optional
            This Nodes AddAction when created in Server, by default None
        target : Node or int or None, optional
            This Nodes AddActions target, by default None
        server : SCServer, optional
            The Server for this Node,
            by default use the SC default server
        """
        self._server = server or sc3nb.SC.get_default().server
        if nodeid in self._server.nodes:
            raise RuntimeError("The __init__ of Node should not be called twice")

        self._state_lock = RLock()
        self._free_event = Event()
        self._on_free_callback = None
        self._started = False
        self._freed = False

        self._nodeid = (
            nodeid if nodeid is not None else self._server.node_ids.allocate(1)[0]
        )
        self._group = None

        self._target_id = Node._get_nodeid(target) if target is not None else None
        if add_action is not None:
            self._add_action = AddAction(add_action)
        else:
            self._add_action = AddAction.TO_HEAD

        self._set_node_attrs(target, add_action)

        # only with node watcher
        self._is_playing = None
        self._is_running = None

        # this is state that we cannot really be sure of
        self._current_controls = {}

        _LOGGER.debug(
            "Adding new %s(%s) to %s", type(self).__name__, self._nodeid, self._server
        )
        self._server.nodes[self._nodeid] = self

    @abstractmethod
    def new(
        self,
        *args,
        add_action: Optional[Union[AddAction, int]] = None,
        target: Optional[Union["Node", int]] = None,
        return_msg: bool = False,
        **kwargs,
    ) -> Union["Node", OSCMessage]:
        """Create a new Node

        Parameters
        ----------
        add_action : AddAction or int, optional
            Where the Node should be added, by default AddAction.TO_HEAD (0)
        target : Node or int, optional
            AddAction target, if None it will be the default group of the server
        """
        with self._state_lock:
            self._started = True
            self._freed = False
            self._free_event.clear()
            self._is_playing = None
            self._is_running = None
            self._set_node_attrs(target=target, add_action=add_action)

    def _get_status_repr(self) -> str:
        status = ""
        if self.started and not self.is_playing:
            status = "s"
        if self.is_playing:
            running_symbol = "~" if self.is_running else "-"
            status = running_symbol
        if self.freed:
            status = "f"
        return status

    def _set_node_attrs(
        self,
        target: Optional[Union["Node", int]] = None,
        add_action: Optional[Union[AddAction, int]] = None,
    ) -> None:
        """Derive Node group from addaction and target

        Parameters
        ----------
        target : int or Node
            Target nodeid or Target Node of this Node's AddAction
        add_action : AddAction
            AddAction of this Node, default AddAction.TO_HEAD (0)
        """
        with self._state_lock:
            _LOGGER.debug(
                "Node attrs before setting: nodeid %s, group %s, add_action %s, target %s",
                self._nodeid,
                self._group,
                self._add_action,
                self._target_id,
            )
            # get target id
            if target is not None:
                self._target_id = Node._get_nodeid(target)
            else:
                if self._target_id is None:
                    self._target_id = self.server.default_group.nodeid

            # get add action
            if add_action is None:
                self._add_action = AddAction.TO_HEAD
            elif isinstance(add_action, AddAction):
                self._add_action = add_action
            else:
                self._add_action = AddAction(add_action)

            # derive group
            if self._add_action in [AddAction.TO_HEAD, AddAction.TO_TAIL]:
                self._group = self._target_id
            elif self._group is None:  # AddAction BEFORE, AFTER or REPLACE
                if isinstance(target, Node):
                    self._group = target.group
                elif target in self._server.nodes:
                    target_node = self._server.nodes[self._target_id]
                    if target_node:
                        self._group = target_node.group
                else:
                    _LOGGER.warning(
                        "Could not derive group of Node, assuming default group"
                    )
                    self._group = self._server.default_group.nodeid
            _LOGGER.debug(
                "Node attrs after setting: nodeid %s, group %s, add_action %s, target %s",
                self._nodeid,
                self._group,
                self._add_action,
                self._target_id,
            )

    @property
    def nodeid(self) -> int:
        """Identifier of node."""
        return self._nodeid

    @property
    def group(self) -> Optional[int]:
        """Identifier of this nodes group."""
        return self._group

    @property
    def server(self) -> "SCServer":
        """The server on which this node is located."""
        return self._server

    @property
    def is_playing(self) -> Optional[bool]:
        """True if this node is playing. None if unkown."""
        return self._is_playing

    @property
    def is_running(self) -> Optional[bool]:
        """True if this node is running. None if unkown."""
        return self._is_running

    @property
    def freed(self) -> bool:
        """True if free was called on this node.

        This is reseted when receiving a /n_go notification
        """
        return self._freed

    @property
    def started(self) -> bool:
        """True if new was called on this node.

        This is reseted when receiving a /n_end notification
        """
        return self._started

    def free(self, return_msg: bool = False) -> Union["Node", OSCMessage]:
        """Free the node with /n_free.

        This will set is_running and is_playing to false.
        Even when the message is returned to mimic the behavior of the SuperCollider Node
        See https://doc.sccode.org/Classes/Node.html#-freeMsg

        Returns
        -------
        Node or OSCMessage
            self for chaining or OSCMessage when return_msg=True
        """
        with self._state_lock:
            self._freed = True
        msg = OSCMessage(NodeCommand.FREE, [self.nodeid])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def run(
        self, on: bool = True, return_msg: bool = False
    ) -> Union["Node", OSCMessage]:
        """Turn node on or off with /n_run.

        Parameters
        ----------
        on : bool
            True for on, False for off, by default True

        Returns
        -------
        Node or OSCMessage
            self for chaining or OSCMessage when return_msg=True
        """
        msg = OSCMessage(NodeCommand.RUN, [self.nodeid, 1 if on else 1])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def set(
        self,
        argument: Union[str, Dict, List],
        *values: Any,
        return_msg: bool = False,
    ) -> Union["Node", OSCMessage]:
        """Set a control value(s) of the node with n_set.

        Parameters
        ----------
        argument : str | dict | list
            if string: name of control argument
            if dict: dict with argument, value pairs
            if list: use list as message content
        value : any, optional
            only used if argument is string, by default None

        Examples
        --------

        >>> synth.set("freq", 400)
        >>> synth.set({"dur": 1, "freq": 400})
        >>> synth.set(["dur", 1, "freq", 400])

        """
        # update cached current_control values
        with self._state_lock:
            msg_params: List[Any] = [self.nodeid]
            if isinstance(argument, dict):
                for arg, val in argument.items():
                    msg_params.append(arg)
                    msg_params.append(val)
                    self._update_control(arg, val)
            elif isinstance(argument, list):
                for arg_idx, arg in enumerate(argument):
                    if isinstance(arg, str):
                        self._update_control(arg, argument[arg_idx + 1])
                msg_params.extend(argument)
            else:
                if len(values) == 1:
                    self._update_control(argument, values[0])
                else:
                    self._update_control(argument, values)
                msg_params.extend([argument] + list(values))
        msg = OSCMessage(NodeCommand.SET, msg_params)
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def _update_control(self, control: str, value: Any) -> None:
        with self._state_lock:
            try:
                val = object.__getattribute__(self, control)
            except AttributeError:
                pass
            else:
                warnings.warn(
                    f"attribute {control}={val} is deleted and recognized as Node Parameter now"
                )
                delattr(self, control)
            if not control.startswith("t_"):
                self._current_controls[control] = value

    def _update_controls(self, controls: Optional[Dict[str, Any]] = None) -> None:
        with self._state_lock:
            if controls is not None:
                for arg, val in controls.items():
                    self._update_control(arg, val)

    def fill(
        self,
        control: Union[str, int],
        num_controls: int,
        value: Any,
        return_msg: bool = False,
    ) -> Union["Node", OSCMessage]:
        """Fill ranges of control values with n_fill.

        Parameters
        ----------
        control : int or string
            control index or name
        num_controls : int
            number of control values to fill
        value : float or int
            value to set
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        msg = OSCMessage(NodeCommand.FILL, [self.nodeid, control, num_controls, value])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def map(
        self, control: Union[str, int], bus: "Bus", return_msg: bool = False
    ) -> Union["Node", OSCMessage]:
        """Map a node's control to read from a bus using /n_map or /n_mapa.

        Parameters
        ----------
        control : int or string
            control index or name
        bus : Bus
            control/audio bus
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        msg_params = [self.nodeid, control, bus.idxs[0]]
        if bus.num_channels > 1:
            map_command = NodeCommand.MAPAN if bus.is_audio_bus() else NodeCommand.MAPN
            msg_params.append(bus.num_channels)
        else:
            map_command = NodeCommand.MAPA if bus.is_audio_bus() else NodeCommand.MAP
        msg = OSCMessage(map_command, msg_params)
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def release(
        self, release_time: Optional[float] = None, return_msg: bool = False
    ) -> Union["Node", OSCMessage]:
        """Set gate as specified.

        https://doc.sccode.org/Classes/Node.html#-release

        Parameters
        ----------
        release_time : float, optional
            amount of time in seconds during which the node will release.
            If set to a value <= 0, the synth will release immediately.
            If None using its Envs normal release stage(s)
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        if release_time is not None:
            release_time = 1 if release_time <= 0 else -1 * (release_time + 1)
        else:
            release_time = 0

        msg = OSCMessage(NodeCommand.SET, [self.nodeid, "gate", release_time])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def query(self) -> Union[SynthInfo, GroupInfo]:
        """Sends an n_query message to the server.

        The answer is send to all clients who have registered via the /notify command.
        Content of answer:

        node ID
        the node's parent group ID
        previous node ID, -1 if no previous node.
        next node ID, -1 if no next node.
        1 if the node is a group, 0 if it is a synth

        if the node is a group:
            ID of the head node, -1 if there is no head node.
            ID of the tail node, -1 if there is no tail node.

        Returns
        -------
        SynthInfo or GroupInfo
            n_info answer. See above for content description
        """
        msg = OSCMessage(NodeCommand.QUERY, [self.nodeid])
        result = self.server.send(msg, bundle=False)
        return self._parse_info(*result)

    def trace(self, return_msg: bool = False) -> Union["Node", OSCMessage]:
        """Trace a node.

        Print out values of the inputs and outputs for one control period.
        If node is a group then print the node IDs and names of each node.

        Parameters
        ----------
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        Node or OSCMessage
            if return_msg else self
        """
        msg = OSCMessage(NodeCommand.TRACE, [self.nodeid])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def move(
        self, add_action: AddAction, another_node: "Node", return_msg: bool = False
    ) -> Union["Node", OSCMessage]:
        """Move this node

        Parameters
        ----------
        add_action : AddAction [TO_HEAD, TO_TAIL, AFTER, BEFORE]
            What add action should be done.
        another_node : Node
            The node which is the target of the add action
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        Node or OSCMessage
            if return_msg this will be the OSCMessage, else self

        Raises
        ------
        ValueError
            If a wrong AddAction was provided
        """
        if add_action == AddAction.REPLACE:
            raise ValueError(
                "add_action needs to be in [TO_HEAD, TO_TAIL, AFTER, BEFORE]"
            )
        msg = OSCMessage(
            NodeCommand.ORDER, [add_action.value, another_node.nodeid, self.nodeid]
        )
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def register(self):
        """Register to be watched."""
        raise NotImplementedError("Currently all nodes are being watched on default")

    def unregister(self):
        """Unregister to stop being watched."""
        raise NotImplementedError("Currently all nodes are being watched on default")

    def on_free(self, func):
        """Callback that is executed when this Synth is freed"""
        self._on_free_callback = func

    def wait(self, timeout: Optional[float] = None) -> None:
        """Wait until this Node is freed

        Raises
        ------
        TimeoutError
            If timeout was provided and wait timed out.
        """
        # TODO check if self._server is in bundling mode,
        # This should probably fail if used inside of Bundler
        if not self._free_event.wait(timeout=timeout):
            raise TimeoutError("Timed out waiting for synth.")

    def _parse_info(
        self,
        nodeid: int,
        group: int,
        prev_nodeid: int,
        next_nodeid: int,
        *rest: Sequence[int],
    ) -> Union[SynthInfo, GroupInfo]:
        assert (
            self.nodeid == nodeid
        ), "nodeids does not match self.nodeid={self.nodeid} != {nodeid}"
        if len(rest) == 1 and rest[0] == 0:  # node is synth
            return SynthInfo._make([nodeid, group, prev_nodeid, next_nodeid])
        else:
            _, head, tail = rest
            return GroupInfo._make(
                [nodeid, group, prev_nodeid, next_nodeid, head, tail]
            )

    def _handle_notification(self, kind: str, info) -> None:
        with self._state_lock:
            if kind == "/n_go":
                self._is_playing = True
                self._is_running = True
                self._started = True
                self._freed = False
            elif kind == "/n_end":
                self._is_playing = False
                self._is_running = False
                self._started = False
                self._freed = True
                self._group = None
                self._free_event.set()
                if self._on_free_callback is not None:
                    self._on_free_callback()
            elif kind == "/n_on":
                self._is_running = True
                self._freed = False
            elif kind == "/n_off":
                self._is_running = False
                self._started = False
            elif kind == "/n_move":
                node_info = self._parse_info(*info)
                self._group = node_info.group
            else:
                raise ValueError(f"Illegal notification kind '{kind}' {info}")
        _LOGGER.debug("Handled %s notification: %s", kind, info)

    def __eq__(self, other):
        return self.nodeid == other.nodeid

    @staticmethod
    def _get_nodeid(value: Union["Node", int]) -> int:
        """Get the corresponding node id

        Parameters
        ----------
        value : Node or int
            If a Node is provided it will get its nodeid
            If a int is provided it will be returned

        Returns
        -------
        int
            nodeid

        Raises
        ------
        ValueError
            When neither Node or int was provided
        """
        if isinstance(value, Node):
            nodeid = value.nodeid
        elif isinstance(value, int):
            nodeid = value
        else:
            raise ValueError("Could not get a node id")
        return nodeid


class Synth(Node):
    """Representation of a Synth on SuperCollider."""

    def __init__(
        self,
        name: Optional[str] = None,
        controls: Dict[str, Any] = None,
        *,
        nodeid: Optional[int] = None,
        new: bool = True,
        add_action: Optional[Union[AddAction, int]] = None,
        target: Optional[Union["Node", int]] = None,
        server: Optional["SCServer"] = None,
    ) -> None:
        """Create a Python representation of a SuperCollider synth.

        Parameters
        ----------
        sc : SC
            sc3nb SuperCollider instance
        name : str, optional
            name of the synth to be created, by default "default"
        controls : dict, optional
            synth control arguments, by default None
        nodeid : int, optional
            ID of the node in SuperCollider, by default sc3nb will create one.
            Can be set to an existing id to create a Python instance of a running Node.
        new : bool, optional
            True if synth should be created on the server, by default True
            Should be False if creating an instance of a running Node.
        add_action : AddAction or int, optional
            Where the Synth should be added, by default AddAction.TO_HEAD (0)
        target : Node or int, optional
            AddAction target, if None it will be the default group of the server

        Raises
        ------
        ValueError
            Raised when synth can't be found via SynthDescLib.global

        Examples
        --------

        >>> scn.Synth(sc, "s1", {"dur": 1, "freq": 400})

        """
        self._server = server or sc3nb.SC.get_default().server
        if nodeid in self._server.nodes:
            self._update_synth_state(name=name, controls=controls)
            if new:
                self.new(controls=controls, add_action=add_action, target=target)
            return

        # attention: this must be the first line. see __setattr__, __getattr__
        self._initialized = False
        super().__init__(
            nodeid=nodeid,
            add_action=add_action,
            target=target,
            server=server,
        )
        with self._state_lock:
            self._name = name or "default"
            self._synth_desc = SynthDef.get_description(self._name)
            if controls is None:
                controls = {}
            self._current_controls = controls

            self._freed = False
            self._started = False

            # attention: this must be after every attribute is set
            self._initialized = True
        if new:
            self.new(
                controls=self._current_controls,
                add_action=self._add_action,
                target=self._target_id,
            )

    def _update_synth_state(self, name: Optional[str], controls: Optional[dict]):
        _LOGGER.debug("Update Synth(%s)", self.nodeid)
        with self._state_lock:
            if name is not None:
                self._name = name
                self._synth_desc = SynthDef.get_description(name)
            self._update_controls(controls)

    @property
    def synth_desc(self) -> Optional[Dict[str, "SynthArgument"]]:
        """A Description of this Synths arguments"""
        with self._state_lock:
            if self._synth_desc is None:
                self._synth_desc = SynthDef.get_description(self._name)
        return self._synth_desc

    @property
    def name(self) -> str:
        """This Synths SynthDef name."""
        return self._name

    @property
    def current_controls(self) -> Dict[str, Any]:
        """This Synth currently cached control arguments."""
        return self._current_controls

    def new(
        self,
        controls: Optional[dict] = None,
        add_action: Optional[Union[AddAction, int]] = None,
        target: Optional[Union[Node, int]] = None,
        *,
        return_msg: bool = False,
    ) -> Union["Synth", OSCMessage]:
        """Creates the synth on the server with s_new.

        Attention: Here you create an identical synth! Same nodeID etc.
        - This will fail if there is already this nodeID on the SuperCollider server!
        """
        with self._state_lock:
            super().new(target=target, add_action=add_action)
            self._update_controls(controls)
            flatten_args = reduce(iconcat, self._current_controls.items(), [])

        msg = OSCMessage(
            SynthCommand.NEW,
            [self._name, self.nodeid, self._add_action.value, self._target_id]
            + flatten_args,
        )
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True, await_reply=False)
        return self

    def get(self, control: str) -> Any:
        """Get a Synth argument

        This will request the value from scsynth with /s_get(n).

        Parameters
        ----------
        control : str
            name of the Synth control argument
        """
        with self._state_lock:
            if (
                self._synth_desc is not None
            ):  # change from synth_desc to self._current_controls
                try:
                    default_value = self._synth_desc[control].default
                except KeyError as error:
                    raise ValueError(
                        f"argument '{control}' not in synth_desc"
                        f"{self._synth_desc.keys()}"
                    ) from error
            else:
                default_value = None
        # if we know they type of the argument and its list we use s_getn
        if default_value is not None and isinstance(default_value, list):
            command = SynthCommand.S_GETN
            msg_params = [self.nodeid, control, len(default_value)]
        else:
            command = SynthCommand.S_GET
            msg_params = [self.nodeid, control]

        msg = OSCMessage(command, msg_params)
        try:
            reply = self.server.send(msg, bundle=False)
        except OSCCommunicationError as osc_error:
            if command in self.server.fails:
                fail = self.server.fails.msg_queues[command].get(timeout=0)
                if f"Node {self.nodeid} not found" in fail:
                    raise RuntimeError(
                        f"Node {self.nodeid} cannot be found"
                    ) from osc_error
            raise
        else:
            if default_value is not None and isinstance(default_value, list):
                nodeid, name, _, *values = reply
                ret_val = list(values)
            else:  # default s_get
                nodeid, name, ret_val = reply
            if self.nodeid == nodeid and name == control:
                self.current_controls[control] = ret_val
                return ret_val
            else:
                raise RuntimeError("Received msg with wrong node id")

    def seti(self, *args):
        """Set part of an arrayed control."""
        raise NotImplementedError()

    def __getattr__(self, name):
        # python will try obj.__getattribute__(name) before this
        if self._initialized:
            with self._state_lock:
                if (
                    name in self._current_controls
                    or self._synth_desc
                    and name in self._synth_desc
                ):
                    return self.get(name)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name, value):
        # First try regular attribute access.
        # This is done similiar in pandas NDFrame.
        try:
            object.__getattribute__(self, name)
            return object.__setattr__(self, name, value)
        except AttributeError:
            pass

        # First time the _initialized and _server is not set
        # and then it is false until Synth instance is done with __init__
        if name in ["_server", "_initialized"] or not self._initialized:
            return super().__setattr__(name, value)
        # if initialized try setting current controls
        with self._state_lock:
            if name in self._current_controls or (
                self._synth_desc and name in self._synth_desc
            ):
                return self.set(name, value)
        warnings.warn(
            f"Setting '{name}' as python attribute and not as Synth Parameter. "
            "SynthDesc is unknown. sclang must be running and knowing this SynthDef "
            "Use set method when using Synths without SynthDesc to set Synth Parameters."
        )
        super().__setattr__(name, value)

    def __repr__(self) -> str:
        status = self._get_status_repr()
        return (
            f"<Synth({self.nodeid}) '{self._name}' {status} {self._current_controls}>"
        )


class Group(Node):
    """Representation of a Group on SuperCollider."""

    def __init__(
        self,
        *,
        nodeid: Optional[int] = None,
        new: bool = True,
        parallel: bool = False,
        add_action: AddAction = AddAction.TO_HEAD,
        target: Optional[Union[Node, int]] = None,
        server: Optional["SCServer"] = None,
    ) -> None:
        """Create a Python representation of a SuperCollider group.

        Parameters
        ----------
        nodeid : int, optional
            ID of the node in SuperCollider, by default sc3nb will create one.
            Can be set to an existing id to create a Python instance of a running Node.
        new : bool, optional
            True if synth should be created on the server, by default True
            Should be False if creating an instance of a running Node.
        parallel : bool, optional
            If True create a parallel group, by default False
        add_action : AddAction or int, optional
            Where the Group should be added, by default AddAction.TO_HEAD (0)
        target : Node or int, optional
            AddAction target, if None it will be the default group of the server
        server : SCServer, optional
            Server instance where this Group is located,
            by default use the SC default server
        """
        self._server = server or sc3nb.SC.get_default().server
        if nodeid in self._server.nodes:
            if new:
                self.new(add_action=add_action, target=target)
            return

        super().__init__(
            nodeid=nodeid,
            add_action=add_action,
            target=target,
            server=server,
        )
        with self._state_lock:
            self._parallel = parallel
            self._children = []

        if new:
            self.new(add_action=self._add_action, target=self._target_id)

    def _update_group_state(
        self,
        children: Optional[Sequence[Node]] = None,
    ) -> None:
        _LOGGER.debug("Update Group(%s)", self.nodeid)
        with self._state_lock:
            if children is not None:
                self._children = children

    def new(
        self,
        add_action=AddAction.TO_HEAD,
        target=None,
        *,
        parallel=None,
        return_msg=False,
    ) -> Union["Group", OSCMessage]:
        """Creates the synth on the server with g_new / p_new.

        Attention: Here you create an identical group! Same nodeID etc.
        - This will fail if there is already this nodeID on the SuperCollider server!

        Parameters
        ----------
        add_action : AddAction or int, optional
            where the group should be added, by default AddAction.TO_HEAD (0)
        target : Node or int, optional
            add action target, by default 1
        parallel : bool, optional
            If True use p_new, by default False
        return_msg : bool, optional
            If ture return the OSCMessage instead of sending it, by default False

        Returns
        -------
        Group
            self
        """
        with self._state_lock:
            super().new(target=target, add_action=add_action)
            if parallel is not None:
                self._parallel = parallel
            new_command = GroupCommand.P_NEW if self._parallel else GroupCommand.G_NEW
        msg = OSCMessage(
            new_command, [self.nodeid, self._add_action.value, self._target_id]
        )
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True, await_reply=False)
        return self

    @property
    def children(self) -> Sequence[Node]:
        """Return this groups children as currently known

        Returns
        -------
        Sequence[Node]
            Sequence of child Nodes (Synths or Groups)
        """
        return self._children

    def move_node_to_head(self, node, return_msg=False):
        """Move node to this groups head with g_head.

        Parameters
        ----------
        node : Node
            node to move
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        Group
            self
        """
        msg = OSCMessage(GroupCommand.HEAD, [self.nodeid, node.nodeid])
        self.server.send(msg, bundle=True)
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def move_node_to_tail(self, node, return_msg=False):
        """Move node to this groups tail with g_tail.

        Parameters
        ----------
        node : Node
            node to move
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        Group
            self
        """
        msg = OSCMessage(GroupCommand.TAIL, [self.nodeid, node.nodeid])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def free_all(self, return_msg=False):
        """Frees all nodes in the group with g_freeAll.

        Parameters
        ----------
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        self._children = []
        msg = OSCMessage(GroupCommand.FREE_ALL, [self.nodeid])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def deep_free(self, return_msg=False):
        """Free all synths in this group and its sub-groups with g_deepFree.

        Sub-groups are not freed.

        Parameters
        ----------
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        with self._state_lock:
            self._children = [c for c in self._children if isinstance(c, Group)]
        msg = OSCMessage(GroupCommand.DEEP_FREE, [self.nodeid])
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def dump_tree(self, post_controls=True, return_msg=False):
        """Posts a representation of this group's node subtree with g_dumpTree.

        Parameters
        ----------
        post_controls : bool, optional
            True for control values, by default False
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Returns
        -------
        OSCMessage
            if return_msg else self
        """
        msg = OSCMessage(
            GroupCommand.DUMP_TREE, [self.nodeid, 1 if post_controls else 0]
        )
        if return_msg:
            return msg
        else:
            self.server.send(msg, bundle=True)
        return self

    def query_tree(self, include_controls=False) -> "Group":
        """Send a g_queryTree message for this group.

        See https://doc.sccode.org/Reference/Server-Command-Reference.html#/g_queryTree for details.

        Parameters
        ----------
        include_controls : bool, optional
            True for control values, by default False

        Returns
        -------
        tuple
            /g_queryTree.reply
        """
        msg = OSCMessage(
            GroupCommand.QUERY_TREE, [self.nodeid, 1 if include_controls else 0]
        )
        _, *nodes_info = self.server.send(msg)
        NodeTree(
            info=nodes_info,
            root_nodeid=self.nodeid,
            controls_included=include_controls,
            start=0,
            server=self.server,
        )
        return self

    def _repr_pretty_(self, printer, cylce):
        status = self._get_status_repr()
        if cylce:
            printer.text(f"Group({self.nodeid}) {status}>")
        else:
            printer.text(f"Group({self.nodeid}) {status} {self._current_controls}")
            with printer.group(2, " children=[", "]"):
                if self._children:
                    printer.breakable()
                    for idx, child in enumerate(self._children):
                        if idx:
                            printer.text(",")
                            printer.breakable()
                        printer.pretty(child)

    def __repr__(self) -> str:
        status = self._get_status_repr()
        return f"<Group({self.nodeid}) {status} {self._current_controls} children={self.children}>"


class NodeTree:
    """Node Tree is a class for parsing /g_queryTree.reply"""

    def __init__(
        self,
        info: Sequence[Any],
        root_nodeid: int,
        controls_included: bool,
        start: int = 0,
        server: Optional["SCServer"] = None,
    ) -> None:
        self.controls_included = controls_included
        self.root_nodeid = root_nodeid
        parsed, self.root = NodeTree.parse_nodes(info, controls_included, start, server)
        assert len(info) == parsed, "Mismatch in nodes info length and parsed info"

    @staticmethod
    def parse_nodes(
        info: Sequence[Any],
        controls_included: bool = True,
        start: int = 0,
        server: Optional["SCServer"] = None,
    ) -> Tuple[int, Node]:
        """Parse Nodes from reply of the /g_queryTree cmd of scsynth.
        This reads the /g_queryTree.reply and creates the corresponding Nodes in Python.
        See https://doc.sccode.org/Reference/Server-Command-Reference.html#/g_queryTree

        Parameters
        ----------
        controls_included : bool
            If True the current control (arg) values for synths will be included
        start : int
            starting position of the parsing, used for recursion, default 0
        info : Sequence[Any]
            /g_queryTree.reply to be parsed.

        Returns
        -------
        Tuple[int, Node]
            postion where the parsing ended, resulting Node
        """
        pos = start + 2
        nodeid, num_children = info[start:pos]
        if num_children < 0:  # -1 children ==> synth
            symbol = info[pos:][0]
            pos += 1
            num_controls = None
            controls = None
            if controls_included:
                num_controls = info[pos:][0]
                pos += 1
                controls_size = 2 * num_controls
                controls_info = info[pos:][:controls_size]
                controls = dict(zip(controls_info[::2], controls_info[1::2]))
                pos += controls_size
            return (
                pos,
                Synth(
                    name=symbol,
                    controls=controls,
                    nodeid=nodeid,
                    new=False,
                    server=server,
                ),
            )
        # num_children >= 0 ==> group
        children = []
        to_parse = num_children
        while to_parse > 0:
            # is group
            pos, node = NodeTree.parse_nodes(info, controls_included, pos, server)
            node._group = nodeid
            children.append(node)
            to_parse -= 1
        group = Group(nodeid=nodeid, new=False, server=server)
        group._update_group_state(children=children)
        return pos, group

    def _repr_pretty_(self, printer, cylce):
        if cylce:
            printer.text(f"NodeTree root={self.root_nodeid}")
        else:
            printer.text("NodeTree root=")
            printer.pretty(self.root)



""" Module for process handling. """

import atexit
import glob
import logging
import os
import platform
import re
import subprocess
import threading
import time
import warnings
from queue import Empty, Queue
from typing import Optional, Sequence

import psutil

_LOGGER = logging.getLogger(__name__)


ANSI_ESCAPE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
ALLOWED_PARENTS = ("scide", "python", "tox")


def find_executable(
    executable: str, search_path: str = None, add_to_path: bool = False
):
    """Looks for executable in os $PATH or specified path

    Parameters
    ----------
    executable : str
        Executable to be found
    search_path : str, optional
        Path at which to look for, by default None
    add_to_path : bool, optional
        Wether to add the provided path to os $PATH or not, by default False

    Returns
    -------
    str
        Full path to executable

    Raises
    ------
    FileNotFoundError
        Raised if executable cannot be found
    """
    paths = os.environ["PATH"].split(os.pathsep)

    if search_path is not None:
        head, tail = os.path.split(search_path)
        if executable == os.path.splitext(tail):
            search_path = head
        if search_path not in os.environ["PATH"] and add_to_path:
            os.environ["PATH"] += os.pathsep + search_path
        paths.insert(0, search_path)

    if platform.system() == "Darwin":
        for directory in ["/Applications/SuperCollider/", "/Applications/"]:
            if executable == "sclang":
                paths.append(directory + "SuperCollider.app/Contents/MacOS/")
            elif executable == "scsynth":
                paths.append(directory + "SuperCollider.app/Contents/Resources/")
    elif platform.system() == "Windows":
        paths.extend(glob.glob("C:/Program Files/SuperCollider-*/"))
    # elif platform.system() == "Linux":
    _LOGGER.debug("Searching executable in paths: %s", paths)
    extlist = [""]
    if os.name == "os2":
        _, ext = os.path.splitext(executable)
        # executable files on OS/2 can have an arbitrary extension, but
        # .exe is automatically appended if no dot is present in the name
        if not ext:
            executable += ".exe"
    elif platform.system() == "Windows":
        pathext = os.environ["PATHEXT"].lower().split(os.pathsep)
        _, ext = os.path.splitext(executable)
        if ext.lower() not in pathext:
            extlist = pathext

    for ext in extlist:
        execname = executable + ext
        for path in paths:
            file = os.path.join(path, execname)
            if os.path.isfile(file):
                return file
    raise FileNotFoundError(
        f"Unable to find '{executable}' executable in {paths} (platform.system={platform.system()})"
    )


def kill_processes(exec_path, allowed_parents: Optional[tuple] = None):
    """Kill processes with the same path for the executable.

    If allowed_parent is provided it will be searched in the names of the
    parent processes of the process with the executable path before terminating.
    If it is found the process won't be killed.

    Parameters
    ----------
    exec_path : str
        path of the executable to kill
    allowed_parent : str, optional
        parents name of processes to keep, by default None
    """
    _LOGGER.debug("Trying to find leftover processes of %s", exec_path)
    for proc in psutil.process_iter(["exe"]):
        if proc.info["exe"] == exec_path:
            parents = proc.parents()
            if allowed_parents and parents:
                parent_names = " ".join(map("".join, map(psutil.Process.name, parents)))
                _LOGGER.debug("Parents cmdlines: %s", parent_names)
                if any(
                    allowed_parent in parent_names for allowed_parent in allowed_parents
                ):
                    continue
            _LOGGER.warning(
                "Found old process. Please exit sc3nb via sc.exit(). \n"
                " Terminating %s because none of"
                " the parents=%s are in allowed_parents=%s"
                " More information can be found in the documentation.",
                proc,
                parents,
                allowed_parents,
            )
            proc.terminate()
            try:
                proc.wait(timeout=0.5)
            except psutil.TimeoutExpired:
                proc.kill()
                warnings.warn(f"Couldn't terminate {proc}")


class ProcessTimeout(Exception):
    """Process Timeout Exception"""

    def __init__(self, executable, output, timeout, expected):
        self.output = output
        self.timeout = timeout
        message = f"Reading of {executable} timed out after {timeout}s"
        if expected:
            message += f' while expecting: "{expected}"'
        super().__init__(message)


class Process:
    """Class for starting a executable and communication with it.

    Parameters
    ----------
    executable : str
        Name of executable to start
    programm_args : Optional[Sequence[str]], optional
        Arguments to program start with Popen, by default None
    executable_path : str, optional
        Path with executalbe, by default system PATH
    console_logging : bool, optional
        Flag for controlling console logging, by default True
    kill_others : bool, optional
        Flag for controlling killing of other executables with the same name.
        This is useful when processes where left over, by default True
    allowed_parents : Sequence[str], optional
        Sequence of parent names that won't be killed
        when kill_others is True, by default None
    """

    def __init__(
        self,
        executable: str,
        programm_args: Optional[Sequence[str]] = None,
        executable_path: str = None,
        console_logging: bool = True,
        kill_others: bool = True,
        allowed_parents: Sequence[str] = None,
    ):
        self.executable = executable
        self.exec_path = find_executable(self.executable, search_path=executable_path)
        self.console_logging = console_logging
        self.popen_args = [self.exec_path]
        if programm_args is not None:
            self.popen_args += programm_args

        # kill other (leftover) subprocesses
        # https://github.com/jupyter/jupyter_client/issues/104
        if kill_others:
            kill_processes(self.exec_path, allowed_parents)

        # starting subprocess
        _LOGGER.info("Popen args: %s", self.popen_args)
        self.popen = subprocess.Popen(
            args=self.popen_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,  # py3.8++ ==> text=True,
            errors="strict",
            bufsize=0,
        )
        # init queue for reading subprocess output queue
        self.output_queue = Queue()
        # output reading worker thread
        self.output_reader_thread = threading.Thread(
            target=self._read_loop, daemon=True
        )
        self.output_reader_thread.start()
        atexit.register(self.kill)

    def _read_loop(self):
        os.write(1, f"[{self.executable} | start reading ]\n".encode())
        for line in iter(self.popen.stdout.readline, ""):
            line = ANSI_ESCAPE.sub("", line)
            if self.console_logging:
                # print to jupyter console...
                os.write(1, f"[{self.executable}]  {line}".encode())
            self.output_queue.put(line)
        os.write(1, f"[{self.executable} | reached EOF ]\n".encode())
        return

    def read(self, expect: Optional[str] = None, timeout: float = 3) -> str:
        """Reads current output from output queue until expect is found

        Parameters
        ----------
        expect : str, optional
            str that we expect to find, by default None
        timeout : float, optional
            timeout in seconds for waiting for output, by default 3

        Returns
        -------
        str
            Output of process.

        Raises
        ------
        ProcessTimeout
            If neither output nor expect is found
        """
        timeout_time = time.time() + timeout
        out = ""
        expect_found = False
        while True:
            if time.time() >= timeout_time:
                raise ProcessTimeout(
                    executable=self.executable,
                    timeout=timeout,
                    output=out,
                    expected=expect,
                )
            try:
                new = self.output_queue.get_nowait()
                if new is not None:
                    out += new
                    if expect is not None and re.search(expect, out):
                        expect_found = True
            except Empty:
                if expect is None or expect_found:
                    return out.strip()
            time.sleep(0.001)

    def empty(self) -> None:
        """Empties output queue."""
        while True:
            try:
                self.output_queue.get_nowait()
            except Empty:
                return

    def write(self, input_str: str) -> None:
        """Send input to process

        Parameters
        ----------
        input_str : str
            Input to be send to process

        Raises
        ------
        RuntimeError
            If writing to process fails
        """
        _LOGGER.debug("Writing: '%s'", input_str)
        try:
            written = self.popen.stdin.write(input_str)
            self.popen.stdin.flush()  # shouldnt be needed as buffering is disabled.
            if written != len(input_str):
                raise RuntimeError(
                    f"Only written {written}/{len(input_str)} of input: '{input_str}'"
                )
        except OSError as error:
            raise RuntimeError("Write to stdin failed") from error

    def kill(self) -> int:
        """Kill the process.

        Returns
        -------
        int
            return code of process
        """
        if self.popen.poll() is None:
            self.popen.kill()
        self.output_reader_thread.join()
        return self.popen.wait()

    def __del__(self):
        self.kill()

    def __repr__(self) -> str:
        returncode = self.popen.returncode
        if returncode is None:
            process_status = f"pid={self.popen.pid}"
        else:
            process_status = f"returncode={returncode}"
        thread_status = "running" if self.output_reader_thread.is_alive() else "died"
        return f"<Process '{self.executable}' ({thread_status}) {process_status}>"



from unittest import TestCase

import numpy as np

from sc3nb.sclang import SCLang, SynthArgument
from tests.conftest import SCBaseTest


class SCLangTest(TestCase):
    def setUp(self) -> None:
        self.sclang = SCLang()

    def tearDown(self) -> None:
        self.sclang.kill()

    def test_started(self):
        self.sclang.start()
        self.assertTrue(self.sclang.started)


class SCLangPersistentTest(SCBaseTest):

    __test__ = True
    start_sclang = True

    def test_auto_load_synthdefs(self):
        self.assertIsNotNone(self.sc.lang.get_synth_description("s1"))
        self.assertIsNotNone(self.sc.lang.get_synth_description("s2"))

    def test_inject_pyvar(self):
        pyvar = "test"
        self.assertEqual(self.sc.lang.cmd("^pyvar.postln", get_result=True), pyvar)

    def test_convert_list_to_sc(self):
        python_list = [1, 2, 3, 4]
        self.assertIn(
            "Array",
            self.sc.lang.cmd(
                """^python_list.class""",
                get_output=True,
                pyvars={"python_list": python_list},
            ),
        )

    def test_convert_nparray_to_sc(self):
        np_array = np.array([1, 2, 3, 4])
        self.assertIn(
            "Array",
            self.sc.lang.cmd(
                """^np_array.class""", get_output=True, pyvars={"np_array": np_array}
            ),
        )

    def test_cmdg_int(self):
        a = 1234
        b = 23452
        sc_val = self.sc.lang.cmdg("""^a+^b""")
        self.assertIsInstance(sc_val, int)
        self.assertEqual(sc_val, a + b)

    def test_cmdg_float(self):
        f = 1234.5
        sc_val = self.sc.lang.cmdg("""^f.squared""")
        self.assertIsInstance(sc_val, float)
        self.assertEqual(sc_val, f ** 2)

    def test_cmdg_str(self):
        soni, fication = "soni", "fication"
        sc_val = self.sc.lang.cmdg("""^soni++^fication""")
        self.assertIsInstance(sc_val, str)
        self.assertEqual(sc_val, soni + fication)

    def test_cmdg_array(self):
        sc_val = self.sc.lang.cmdg("""Array.interpolation(11, 1.0, 2.0)""")
        self.assertIsInstance(sc_val, list)
        self.assertTrue(np.allclose(sc_val, np.linspace(1.0, 2.0, 11)))

    def test_cmdg_list(self):
        pylist = [[1]]
        sc_val = self.sc.lang.cmdg(f"""{pylist.__repr__()}""")
        self.assertIsInstance(sc_val, list)
        self.assertEqual(sc_val, pylist)

        pylist = [[[1]]]
        sc_val = self.sc.lang.cmdg(f"""{pylist.__repr__()}""")
        self.assertIsInstance(sc_val, list)
        self.assertEqual(sc_val, pylist)

        pylist = [[1], [1]]
        sc_val = self.sc.lang.cmdg(f"""{pylist.__repr__()}""")
        self.assertIsInstance(sc_val, list)
        self.assertEqual(sc_val, pylist)

        pylist = [
            [1.0, 1.0],
            [1.100000023841858, 1.100000023841858],
            [1.2000000476837158, 1.2000000476837158],
            [1.2999999523162842, 1.2999999523162842],
            [1.399999976158142, 1.399999976158142],
            [1.5, 1.5],
            [1.600000023841858, 1.600000023841858],
            [1.7000000476837158, 1.7000000476837158],
            [1.7999999523162842, 1.7999999523162842],
            [1.899999976158142, 1.899999976158142],
        ]
        sc_val = self.sc.lang.cmdg(f"""{pylist.__repr__()}""")
        self.assertIsInstance(sc_val, list)
        self.assertEqual(sc_val, pylist)

    def test_get_synth_desc(self):
        expected_synth_desc = {
            "out": SynthArgument("out", "scalar", 0.0),
            "freq": SynthArgument("freq", "control", 440.0),
            "amp": SynthArgument("amp", "control", 0.10000000149011612),
            "pan": SynthArgument("pan", "control", 0.0),
            "gate": SynthArgument("gate", "control", 1.0),
        }
        synth_desc = self.sc.lang.get_synth_description("default")
        self.assertEqual(synth_desc, expected_synth_desc)



"""Collection of helper functions for the user"""
from typing import Optional, Union

import numpy as np


def linlin(
    value: Union[float, np.ndarray],
    x1: float,
    x2: float,
    y1: float,
    y2: float,
    clip: Optional[str] = None,
) -> Union[float, np.ndarray]:
    """Map value linearly so that [x1, x2] is mapped to [y1, y2]

    linlin is implemented in analogy to the SC3 linlin, yet this
    function extrapolates by default.
    A frequently used invocation is with x1 < x2, i.e. thinking
    of them as a range [x1,x2]

    Parameters
    ----------
    value : float or np.ndarray
        value(s) to be mapped
    x1 : float
        source value 1
    x2 : float
        source value 2
    y1 : float
        destination value to be reached for value == x1
    y2 : float
        destination value to be reached for value == x2
    clip: None or string
        None extrapolates, "min" or "max" clip at floor resp. ceiling
        of the destination range, any other value defaults to "minmax",
        i.e. it clips on both sides.

    Returns
    -------
    float or np.ndarray
        the mapping result
    """
    z = (value - x1) / (x2 - x1) * (y2 - y1) + y1
    if clip is None:
        return z
    if y1 > y2:
        x1, x2, y1, y2 = x2, x1, y2, y1
    if clip == "max":
        return np.minimum(z, y2)
    elif clip == "min":
        return np.maximum(z, y1)
    else:  # imply clip to be "minmax"
        return np.minimum(np.maximum(z, y1), y2)


def clip(
    value: float, minimum: float = -float("inf"), maximum: float = float("inf")
) -> float:
    """Clips a value to a certain range

    Parameters
    ----------
    value : float
        Value to clip
    minimum : float, optional
        Minimum output value, by default -float("inf")
    maximum : float, optional
        Maximum output value, by default float("inf")

    Returns
    -------
    float
        clipped value
    """
    if value < minimum:
        return minimum
    if value > maximum:
        return maximum
    return value


def midicps(midi_note: float) -> float:
    """Convert MIDI note to cycles per second

    Parameters
    ----------
    m : float
        midi note

    Returns
    -------
    float
        corresponding cycles per seconds
    """
    return 440.0 * 2 ** ((midi_note - 69) / 12.0)


def cpsmidi(cps: float) -> float:
    """Convert cycles per second to MIDI note

    Parameters
    ----------
    cps : float
        cycles per second

    Returns
    -------
    float
        corresponding MIDI note
    """
    return 69 + 12 * np.log2(cps / 440.0)


def dbamp(decibels: float) -> float:
    """Convert a decibels to a linear amplitude.

    Parameters
    ----------
    decibels : float
        Decibel value to convert

    Returns
    -------
    float
        Corresponding linear amplitude
    """
    return 10 ** (decibels / 20.0)


def ampdb(amp: float) -> float:
    """Convert a linear amplitude to decibels.

    Parameters
    ----------
    amp : float
        Linear amplitude to convert

    Returns
    -------
    float
        Corresponding decibels
    """
    return 20 * np.log10(amp)



"""Module for creating SuperCollider OSC files that can be used for non-realtime synthesis

`SuperCollider Guide - Non-Realtime Synthesis <http://doc.sccode.org/Guides/Non-Realtime-Synthesis.html>`_

"""
import os
import subprocess
import warnings
from typing import Dict, List, Optional, Union

from pythonosc.osc_bundle import OscBundle
from pythonosc.osc_bundle_builder import OscBundleBuilder
from pythonosc.parsing import osc_types

from sc3nb.osc.osc_communication import OSCMessage, convert_to_sc3nb_osc
from sc3nb.process_handling import find_executable
from sc3nb.sc_objects.server import ServerOptions

NRT_SERVER_OPTIONS = "-u 57110 -a 1024 -i 2 -o 2 -R 0 -l 1".split(" ")


class Score:
    @classmethod
    def load_file(
        cls, path: Union[str, bytes, os.PathLike]
    ) -> Dict[float, List[OSCMessage]]:
        """Load a OSC file into a dict.

        Parameters
        ----------
        path : Union[str, bytes, os.PathLike]
            Path of the OSC file.

        Returns
        -------
        Dict[float, List[OSCMessage]]
            dict with time tag as keys and lists of OSCMessages as values.
        """
        with open(path, "rb") as file:
            dgram = file.read()
        index = 0
        builder = OscBundleBuilder(0)
        while dgram[index:]:
            size, index = osc_types.get_int(dgram, index)
            builder.add_content(OscBundle(dgram[index : index + size]))
            index += size
        #  TODO rework that times don't get smashed
        warnings.warn(
            "This method currently destroys the time tag if they happend in the past."
        )
        return convert_to_sc3nb_osc(builder.build()).messages()

    @classmethod
    def write_file(
        cls,
        messages: Dict[float, List[OSCMessage]],
        path: Union[str, bytes, os.PathLike],
        tempo: float = 1,
    ):
        """Write this score as binary OSC file for NRT synthesis.

        Parameters
        ----------
        messages : Dict[float, List[OSCMessage]]
            Dict with times as key and lists of OSC messages as values
        path : Union[str, bytes, os.PathLike]
            output path for the binary OSC file
        tempo : float
            Times will be multiplied by 1/tempo
        """
        tempo_factor = tempo / 1
        with open(path, "wb") as file:
            msg = None
            for timetag, msgs in messages.items():
                builder = OscBundleBuilder((timetag * tempo_factor) - 2208988800.0)
                for msg in msgs:
                    builder.add_content(msg.to_pythonosc())
                dgram = builder.build().dgram
                file.write(osc_types.write_int(len(dgram)))
                file.write(dgram)
            if msg and (msg.address != "/c_set" or msg.parameters != [0, 0]):
                warnings.warn(
                    "Missing /c_set [0, 0] at the end of the messages. "
                    "Recording will stop with last timetag"
                )

    @classmethod
    def record_nrt(
        cls,
        messages: Dict[float, List[OSCMessage]],
        osc_path: str,
        out_file: str,
        in_file: Optional[str] = None,
        sample_rate: int = 44100,
        header_format: str = "AIFF",
        sample_format: str = "int16",
        options: Optional[ServerOptions] = None,
    ):
        """Write an OSC file from the messages and wri

        Parameters
        ----------
        messages : Dict[float, List[OSCMessage]]
            Dict with times as key and lists of OSC messages as values.
        osc_path : str
            Path of the binary OSC file.
        out_file : str
            Path of the resulting sound file.
        in_file : Optional[str], optional
            Path of input soundfile, by default None.
        sample_rate : int, optional
            sample rate for synthesis, by default 44100.
        header_format : str, optional
            header format of the output file, by default "AIFF".
        sample_format : str, optional
            sample format of the output file, by default "int16".
        options : Optional[ServerOptions], optional
            instance of server options to specify server options, by default None

        Returns
        -------
        subprocess.CompletedProcess
            Completed scsynth non-realtime process.
        """
        cls.write_file(messages, osc_path)
        in_file = in_file or "_"
        args = [
            find_executable("scsynth"),
            "-N",
            osc_path,
            in_file,
            out_file,
            str(sample_rate),
            header_format,
            sample_format,
        ]
        if options:
            args.extend(options.options)
        completed_process = subprocess.run(
            args=args,
            check=True,
            universal_newlines=True,  # py>=3.7 text=True
            stdout=subprocess.PIPE,  # py>=3.7 capture_output=True
            stderr=subprocess.PIPE,
        )
        print(completed_process.stdout)
        print(completed_process.stderr)
        return completed_process



import numpy as np

from sc3nb.sc_objects.buffer import Buffer
from tests.conftest import SCBaseTest


class BufferTest(SCBaseTest):

    __test__ = True

    def setUp(self) -> None:
        self.buffer = Buffer()

    def test_load_data(self):
        bufdata = np.random.rand(30000, 1)
        self.buffer.load_data(bufdata)
        self.assertTrue(np.allclose(self.buffer.to_array(), bufdata))



# Minimal setup.py to allow editable installs via pip intstall --editable .
# See https://setuptools.readthedocs.io/en/latest/userguide/quickstart.html#development-mode
import setuptools

setuptools.setup(
    use_scm_version=True,
)



"""Module for recording"""

from enum import Enum, unique
from typing import Optional, Union

from sc3nb.sc import SC
from sc3nb.sc_objects.buffer import Buffer
from sc3nb.sc_objects.node import AddAction, Node, Synth
from sc3nb.sc_objects.server import SCServer
from sc3nb.sc_objects.synthdef import SynthDef


@unique
class RecorderState(Enum):
    """Different States"""

    UNPREPARED = "UNPREPARED"
    PREPARED = "PREPARED"
    RECORDING = "RECORDING"
    PAUSED = "PAUSED"


class Recorder:
    """Allows to record audio easily."""

    # TODO rec_header, rec_format with Literal type (py3.8) from Buffer
    def __init__(
        self,
        path: str = "record.wav",
        nr_channels: int = 2,
        rec_header: str = "wav",
        rec_format: str = "int16",
        bufsize: int = 65536,
        server: Optional[SCServer] = None,
    ):
        """Create and prepare a recorder.

        Parameters
        ----------
        path : str, optional
            path of recording file, by default "record.wav"
        nr_channels : int, optional
            Number of channels, by default 2
        rec_header : str, optional
            File format, by default "wav"
        rec_format : str, optional
            Recording resolution, by default "int16"
        bufsize : int, optional
            size of buffer, by default 65536
        server : SCServer, optional
            server used for recording,
            by default use the SC default server
        """
        self._state = RecorderState.UNPREPARED
        self._server = server or SC.get_default().server
        self._record_buffer = Buffer(server=self._server)
        self._record_synth: Optional[Synth] = None
        self.prepare(path, nr_channels, rec_header, rec_format, bufsize)

    def prepare(
        self,
        path: str = "record.wav",
        nr_channels: int = 2,
        rec_header: str = "wav",
        rec_format: str = "int16",
        bufsize: int = 65536,
    ):
        """Pepare the recorder.

        Parameters
        ----------
        path : str, optional
            path of recording file, by default "record.wav"
        nr_channels : int, optional
            Number of channels, by default 2
        rec_header : str, optional
            File format, by default "wav"
        rec_format : str, optional
            Recording resolution, by default "int16"
        bufsize : int, optional
            size of buffer, by default 65536

        Raises
        ------
        RuntimeError
            When Recorder does not needs to be prepared.
        """
        if self._state != RecorderState.UNPREPARED:
            raise RuntimeError(
                f"Recorder state must be UNPREPARED but is {self._state}"
            )
        # prepare buffer
        self._record_buffer.alloc(bufsize, channels=nr_channels)
        self._record_buffer.write(
            path=path,
            header=rec_header,
            sample=rec_format,
            num_frames=0,
            starting_frame=0,
            leave_open=True,
        )
        self._rec_id = self._record_buffer.bufnum
        # TODO we could prepare the synthDef beforehand and just use the right one here.
        # This would allow Recordings without sclang
        self._synth_def = SynthDef(
            f"sc3nb_recording_{self._rec_id}",
            r"""{ |bus, bufnum, duration|
            var tick = Impulse.kr(1);
            var timer = PulseCount.kr(tick) - 1;
            Line.kr(0, 0, duration, doneAction: if(duration <= 0, 0, 2));
            SendReply.kr(tick, '/recordingDuration', timer, ^rec_id);
            DiskOut.ar(bufnum, In.ar(bus, ^nr_channels))
        }""",
        )
        self._synth_name = self._synth_def.add(
            pyvars={"rec_id": self._rec_id, "nr_channels": nr_channels}
        )
        self._state = RecorderState.PREPARED

    def start(
        self,
        timetag: float = 0,
        duration: Optional[float] = None,
        node: Union[Node, int] = 0,
        bus: int = 0,
    ):
        """Start the recording.

        Parameters
        ----------
        timetag : float, by default 0 (immediately)
            Time (or time offset when <1e6) to start
        duration : float, optional
            Length of the recording, by default until stopped.
        node : Union[Node, int], optional
            Node that should be recorded, by default 0
        bus : int, by default 0
            Bus that should be recorded

        Raises
        ------
        RuntimeError
            When trying to start a recording unprepared.
        """
        if self._state != RecorderState.PREPARED:
            raise RuntimeError(f"Recorder state must be PREPARED but is {self._state}")
        args = dict(bus=bus, duration=duration or -1, bufnum=self._record_buffer.bufnum)

        with self._server.bundler(timetag=timetag):
            self._record_synth = Synth(
                self._synth_name,
                controls=args,
                server=self._server,
                target=node,
                add_action=AddAction.TO_TAIL,
            )
        self._state = RecorderState.RECORDING

    def pause(self, timetag: float = 0):
        """Pause the recording.

        Parameters
        ----------
        timetag : float, by default 0 (immediately)
            Time (or time offset when <1e6) to pause

        Raises
        ------
        RuntimeError
            When trying to pause if not recording.
        """
        if self._state != RecorderState.RECORDING or self._record_synth is None:
            raise RuntimeError(f"Recorder state must be RECORDING but is {self._state}")
        with self._server.bundler(timetag=timetag):
            self._record_synth.run(False)
        self._state = RecorderState.PAUSED

    def resume(self, timetag: float = 0):
        """Resume the recording

        Parameters
        ----------
        timetag : float, by default 0 (immediately)
            Time (or time offset when <1e6) to resume

        Raises
        ------
        RuntimeError
            When trying to resume if not paused.
        """
        if self._state != RecorderState.PAUSED or self._record_synth is None:
            raise RuntimeError(f"Recorder state must be PAUSED but is {self._state}")
        with self._server.bundler(timetag=timetag):
            self._record_synth.run(True)
        self._state = RecorderState.RECORDING

    def stop(self, timetag: float = 0):
        """Stop the recording.

        Parameters
        ----------
        timetag : float, by default 0 (immediately)
            Time (or time offset when <1e6) to stop

        Raises
        ------
        RuntimeError
            When trying to stop if not started.
        """
        if (
            self._state not in [RecorderState.RECORDING, RecorderState.PAUSED]
            or self._record_synth is None
        ):
            raise RuntimeError(
                f"Recorder state must be RECORDING or PAUSED but is {self._state}"
            )

        with self._server.bundler(timetag=timetag):
            self._record_synth.free()
            self._record_synth = None
            self._record_buffer.close()
        self._state = RecorderState.UNPREPARED

    def __repr__(self) -> str:
        return f"<Recorder [{self._state.value}]>"

    def __del__(self):
        try:
            self.stop()
        except RuntimeError:
            pass
        self._record_buffer.free()



import time
import warnings

from sc3nb.sc_objects.node import AddAction, Group, GroupInfo, Synth
from tests.conftest import SCBaseTest


class GroupTest(SCBaseTest):

    __test__ = True

    def setUp(self) -> None:
        with self.assertRaises(RuntimeError):
            GroupTest.sc.lang
        warnings.simplefilter("always", UserWarning)
        self.custom_nodeid = 1010
        self.group = Group(nodeid=self.custom_nodeid)
        GroupTest.sc.server.sync()

    def tearDown(self) -> None:
        nodeid = self.group.nodeid
        self.assertIn(nodeid, self.sc.server.nodes)
        self.group.free()
        self.group.wait(timeout=1)
        del self.group  # make sure that group is deleted from registry
        t0 = time.time()
        while nodeid in self.sc.server.nodes:
            time.sleep(0.005)
            if time.time() - t0 > 0.2:
                self.fail("NodeID is still in server.nodes")
        self.assertNotIn(nodeid, self.sc.server.nodes)
        with self.assertRaises(KeyError):
            del self.sc.server.nodes[nodeid]

    def test_node_registry(self):
        copy1 = Group(nodeid=self.group.nodeid, new=False)
        copy2 = Group(nodeid=self.custom_nodeid, new=False)
        self.assertIs(self.group, copy1)
        self.assertIs(self.group, copy2)
        self.assertIs(copy1, copy2)
        del copy1, copy2

    def test_setattr(self):
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's2' is unknown", msg="SynthDesc seems to be known"
        ):
            synth1 = Synth("s2", controls={"amp": 0.0}, target=self.group)
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's2' is unknown", msg="SynthDesc seems to be known"
        ):
            synth2 = Synth("s2", controls={"amp": 0.0}, target=self.group)
        GroupTest.sc.server.sync()
        cmd_args = {"pan": -1.0, "num": 1}
        for name, value in cmd_args.items():
            self.group.set(name, value)
        GroupTest.sc.server.sync()
        self.assertEqual(synth1.group, self.group.nodeid)
        self.assertEqual(synth2.group, self.group.nodeid)
        for name, value in cmd_args.items():
            self.assertAlmostEqual(synth1.get(name), value)
            self.assertAlmostEqual(synth2.get(name), value)

    # TODO does only trigger a warning if the new group is not the old/same group
    # This seems like a SuperCollider Problem
    def test_new_warning(self):
        with self.assertLogs(level="WARNING") as log:
            self.group.new(target=0)
            time.sleep(0.1)
        self.assertTrue("duplicate node ID" in log.output[-1])

    def test_query(self):
        query_result = self.group.query()
        self.assertIsInstance(query_result, GroupInfo)
        self.assertEqual(query_result.nodeid, self.custom_nodeid)
        self.assertEqual(query_result.group, GroupTest.sc.server.default_group.nodeid)
        self.assertEqual(query_result.prev_nodeid, -1)
        self.assertEqual(query_result.next_nodeid, -1)
        self.assertEqual(query_result.tail, -1)
        self.assertEqual(query_result.head, -1)



"""Module for managing Server related stuff."""
import atexit
import logging
import warnings
from enum import Enum, unique
from queue import Empty
from random import randint
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple
from weakref import WeakValueDictionary

from sc3nb.osc.osc_communication import (
    MessageQueue,
    MessageQueueCollection,
    OSCCommunication,
    OSCCommunicationError,
    OSCMessage,
)
from sc3nb.osc.parsing import preprocess_return
from sc3nb.process_handling import ALLOWED_PARENTS, Process, ProcessTimeout
from sc3nb.sc_objects.allocators import Allocator, BlockAllocator, NodeAllocator
from sc3nb.sc_objects.buffer import BufferCommand, BufferReply
from sc3nb.sc_objects.bus import Bus, BusRate, ControlBusCommand
from sc3nb.sc_objects.node import (
    Group,
    GroupCommand,
    GroupReply,
    Node,
    NodeCommand,
    NodeReply,
    SynthCommand,
)
from sc3nb.sc_objects.synthdef import SynthDef, SynthDefinitionCommand
from sc3nb.sc_objects.volume import Volume

_LOGGER = logging.getLogger(__name__)


@unique
class MasterControlReply(str, Enum):
    """Reply addresses of the Master Control Commands."""

    VERSION_REPLY = "/version.reply"
    SYNCED = "/synced"
    STATUS_REPLY = "/status.reply"


@unique
class MasterControlCommand(str, Enum):
    """Master Control commands of scsynth."""

    DUMP_OSC = "/dumpOSC"
    STATUS = "/status"
    VERSION = "/version"
    CLEAR_SCHED = "/clearSched"
    NOTIFY = "/notify"
    QUIT = "/quit"
    SYNC = "/sync"


@unique
class ReplyAddress(str, Enum):
    """Specific reply addresses."""

    WILDCARD_ADDR = "/*"
    FAIL_ADDR = "/fail"
    DONE_ADDR = "/done"
    RETURN_ADDR = "/return"


ASYNC_CMDS = [
    # Master
    MasterControlCommand.QUIT,
    MasterControlCommand.NOTIFY,
    # Synth Def load SynthDefs
    SynthDefinitionCommand.RECV,
    SynthDefinitionCommand.LOAD,
    SynthDefinitionCommand.LOAD_DIR,
    # Buffer Commands
    BufferCommand.ALLOC,
    BufferCommand.ALLOC_READ,
    BufferCommand.ALLOC_READ_CHANNEL,
    BufferCommand.READ,
    BufferCommand.READ_CHANNEL,
    BufferCommand.WRITE,
    BufferCommand.FREE,
    BufferCommand.ZERO,
    BufferCommand.GEN,
    BufferCommand.CLOSE,
]

CMD_PAIRS = {
    # Master
    MasterControlCommand.STATUS: MasterControlReply.STATUS_REPLY,
    MasterControlCommand.SYNC: MasterControlReply.SYNCED,
    MasterControlCommand.VERSION: MasterControlReply.VERSION_REPLY,
    # Synth Commands
    SynthCommand.S_GET: NodeCommand.SET,
    SynthCommand.S_GETN: NodeCommand.SETN,
    # Group Commands
    GroupCommand.QUERY_TREE: GroupReply.QUERY_TREE_REPLY,
    # Node Commands
    NodeCommand.QUERY: NodeReply.INFO,
    # Buffer Commands
    BufferCommand.QUERY: BufferReply.INFO,
    BufferCommand.GET: BufferCommand.SET,
    BufferCommand.GETN: BufferCommand.SETN,
    # Control Bus Commands
    ControlBusCommand.GET: ControlBusCommand.SET,
    ControlBusCommand.GETN: ControlBusCommand.SETN,
}

LOCALHOST = "127.0.0.1"

SC3NB_SERVER_CLIENT_ID = 1
SC3NB_DEFAULT_PORT = 57130
SCSYNTH_DEFAULT_PORT = 57110
SC3_SERVER_NAME = "scsynth"


class ServerStatus(NamedTuple):
    """Information about the status of the Server program"""

    num_ugens: int
    num_synths: int
    num_groups: int
    num_synthdefs: int
    avg_cpu: float
    peak_cpu: float
    nominal_sr: float
    actual_sr: float


class ServerVersion(NamedTuple):
    """Information about the version of the Server program"""

    name: str
    major_version: int
    minor_version: int
    patch_version: str
    git_branch: str
    commit: str


class ServerOptions:
    """Options for the SuperCollider audio server

    This allows the encapsulation and handling of the command line server options.
    """

    def __init__(
        self,
        udp_port: int = SCSYNTH_DEFAULT_PORT,
        max_logins: int = 6,
        num_input_buses: int = 2,
        num_output_buses: int = 2,
        num_audio_buses: int = 1024,
        num_control_buses: int = 4096,
        num_sample_buffers: int = 1024,
        publish_rendezvous: bool = False,
        block_size: Optional[int] = None,
        hardware_buffer_size: Optional[int] = None,
        hardware_sample_size: Optional[int] = None,
        hardware_input_device: Optional[str] = None,
        hardware_output_device: Optional[str] = None,
        other_options: Optional[Sequence[str]] = None,
    ):
        # options as sequence as wanted by subprocess.Popen
        self.options = []

        # UDP port
        self.udp_port = udp_port
        self.options += ["-u", f"{self.udp_port}"]

        # max logins
        if 3 <= max_logins <= 32:
            self.max_logins = max_logins
        else:
            # max_logins must be between 3 (sc3nb, sclang (sc3nb), sclang (scide)) and 32
            # see https://scsynth.org/t/how-do-i-connect-sclang-to-an-already-running-server/2498/6
            raise ValueError("max logins must be between 3 and 32")
        self.options += ["-l", f"{self.max_logins}"]

        # audio bus options
        self.num_input_buses = num_input_buses
        self.options += ["-i", f"{self.num_input_buses}"]
        self.num_output_buses = num_output_buses
        self.options += ["-o", f"{self.num_output_buses}"]
        if num_audio_buses < num_input_buses + num_output_buses:
            raise ValueError(
                f"You need at least {num_input_buses + num_output_buses} audio buses"
            )
        self.num_audio_buses = num_audio_buses
        self.options += ["-a", f"{self.num_audio_buses}"]

        self.num_control_buses = num_control_buses
        self.options += ["-c", f"{self.num_control_buses}"]

        self.num_sample_buffers = num_sample_buffers
        self.options += ["-b", f"{self.num_sample_buffers}"]

        # publish to Rendezvous
        self.publish_rendezvous = 1 if publish_rendezvous else 0
        self.options += ["-R", f"{self.publish_rendezvous}"]

        if block_size is not None:
            self.block_size = block_size
            self.options += ["-z", f"{self.block_size}"]

        if hardware_buffer_size is not None:
            self.hardware_buffer_size = hardware_buffer_size
            self.options += ["-Z", f"{self.hardware_buffer_size}"]

        if hardware_sample_size is not None:
            self.hardware_sample_size = hardware_sample_size
            self.options += ["-S", f"{self.hardware_sample_size}"]

        # hardware in/out device
        if not hardware_input_device:
            self.hardware_input_device = ""
        else:
            self.hardware_input_device = hardware_input_device
        if not hardware_output_device:
            self.hardware_output_device = ""
        else:
            self.hardware_output_device = hardware_output_device
        if hardware_input_device or hardware_output_device:
            self.options += [
                "-H",
                f"{self.hardware_input_device} {self.hardware_output_device}".strip(),
            ]

        # misc. options
        self.other_options = other_options
        if self.other_options:
            self.options += self.other_options

    @property
    def first_private_bus(self) -> int:
        """The first audio bus after input and output buses"""
        return self.num_output_buses + self.num_input_buses

    @property
    def num_private_buses(self) -> int:
        """Number of audio buses besides input and output buses"""
        return self.num_audio_buses - (self.num_output_buses + self.num_input_buses)

    def __repr__(self):
        return f"<ServerOptions {self.options}>"


class NodeWatcher:
    """The NodeWatcher is used to handle Node Notifications.

    Parameters
    ----------
    server : SCServer
        Server belonging to the notifications
    """

    def __init__(self, server: "SCServer"):
        # self.nodes: Set[int] = set()
        self._server = server
        self.notification_addresses = ["/n_go", "/n_end", "/n_off", "/n_on", "/n_move"]

    # def register(self, node):
    #     self.nodes.add(node.nodeid)

    # def unregister(self, node):
    #     self.nodes.remove(node.nodeid)

    def handle_notification(self, *args):
        """Handle a Notification"""
        kind, *info = args
        nodeid, _, _, _, *rest = info
        # if nodeid in self.nodes:  # check if node is registered
        _LOGGER.debug("Handling %s notification for %s: %s", kind, nodeid, info)
        try:
            node = self._server.nodes[nodeid]
            if node is None:
                raise KeyError("weakref is None")
        except KeyError:
            # we could create the Node here
            # but it wouldn't be refered afterwards and then deleted anyway
            pass
        else:
            node._handle_notification(kind, info)


class SCServer(OSCCommunication):
    """SuperCollider audio server representaion.

    Parameters
    ----------
    options : Optional[ServerOptions], optional
        Options used to start the local server, by default None
    """

    def __init__(self, options: Optional[ServerOptions] = None):
        # process
        if options is None:
            self.options = ServerOptions()
            _LOGGER.debug("Using default server options %s", self.options)
        else:
            self.options = options
            _LOGGER.debug("Using custom server options %s", self.options)

        # counter for nextNodeID
        self._num_node_ids: int = 0

        self.process: Optional[Process] = None
        self._programm_name = SC3_SERVER_NAME

        self._client_id: int = SC3NB_SERVER_CLIENT_ID
        self._scsynth_address = LOCALHOST
        self._scsynth_port = self.options.udp_port
        self._max_logins = self.options.max_logins

        self._server_running: bool = False
        self._has_booted: bool = False

        self.latency: float = 0.0

        self._init_osc_communication()

        # node managing
        self.nodes: WeakValueDictionary[int, Node] = WeakValueDictionary()

        self.node_watcher = NodeWatcher(self)
        for address in self.node_watcher.notification_addresses:
            self._osc_server.dispatcher.map(
                address, self.node_watcher.handle_notification
            )

        self.node_ids: Optional[Allocator] = None
        self.buffer_ids: Optional[BlockAllocator] = None
        self.control_bus_ids: Optional[BlockAllocator] = None
        self.audio_bus_ids: Optional[BlockAllocator] = None

        self._root_node = Group(nodeid=0, new=False, target=0, server=self)
        self._default_groups: Dict[int, Group] = {}
        self._is_local: bool = False

        self._output_bus = Bus(
            rate=BusRate.AUDIO,
            num_channels=self.options.num_output_buses,
            index=0,
            server=self,
        )
        self._input_bus = Bus(
            rate=BusRate.AUDIO,
            num_channels=self.options.num_input_buses,
            index=self.options.num_output_buses,
            server=self,
        )

        self._server_init_hooks: List[Tuple[Callable[..., None], Any, Any]] = []

        self._volume = Volume(self)

        atexit.register(self.quit)

    def boot(
        self,
        scsynth_path: Optional[str] = None,
        timeout: float = 5,
        console_logging: bool = True,
        with_blip: bool = True,
        kill_others: bool = True,
        allowed_parents: Sequence[str] = ALLOWED_PARENTS,
    ):
        """Start the Server process.

        Parameters
        ----------
        scsynth_path : str, optional
            Path of scscynth executable, by default None
        timeout : float, optional
            Timeout for starting the executable, by default 5
        console_logging : bool, optional
            If True write process output to console, by default True
        with_blip : bool, optional
            make a sound when booted, by default True
        kill_others : bool
            kill other SuperCollider server processes.
        allowed_parents : Sequence[str], optional
            Names of parents that are allowed for other instances of
            scsynth processes that won't be killed, by default ALLOWED_PARENTS

        Raises
        ------
        ValueError
            If UDP port specified in options is already used
        ProcessTimeout
            If the process fails to start.
        """
        if self._has_booted:
            warnings.warn("already booted")
            return
        print("Booting SuperCollider Server... ", end="")
        self._is_local = True
        self._scsynth_address = LOCALHOST
        self._scsynth_port = self.options.udp_port
        self._max_logins = self.options.max_logins
        self.process = Process(
            executable=self._programm_name,
            programm_args=self.options.options,
            executable_path=scsynth_path,
            console_logging=console_logging,
            kill_others=kill_others,
            allowed_parents=allowed_parents,
        )
        try:
            self.process.read(expect="SuperCollider 3 server ready.", timeout=timeout)
        except ProcessTimeout as process_timeout:
            if ("Exception in World_OpenUDP" in process_timeout.output) or (
                "ERROR: failed to open UDP socket" in process_timeout.output
            ):
                self.process.kill()
                self.process = None
                print(
                    f"\nSuperCollider Server port {self.options.udp_port} already used."
                )
                if self.options.udp_port != SCSYNTH_DEFAULT_PORT:
                    raise ValueError(
                        f"The specified UDP port {self.options.udp_port} is already used"
                    ) from process_timeout
                print("Trying to connect.")
                self.remote(
                    self._scsynth_address, self._scsynth_port, with_blip=with_blip
                )
            else:
                print("Failed booting SuperCollider Server.")
                raise process_timeout
        else:
            self.init(with_blip)
            self._has_booted = True

    def init(self, with_blip: bool = True):
        """Initialize the server.

        This adds allocators, loads SynthDefs, send default Groups etc.

        Parameters
        ----------
        with_blip : bool, optional
            make a sound when initialized, by default True
        """
        self._init_osc_communication()

        # notify the supercollider server about us
        self.add_receiver(
            self._programm_name, self._scsynth_address, self._scsynth_port
        )
        self.notify()

        # init allocators
        self.node_ids = NodeAllocator(self.client_id)

        buffers_per_user = int(self.options.num_sample_buffers / self.max_logins)
        buffer_id_offset = self.client_id * buffers_per_user
        self.buffer_ids = BlockAllocator(buffers_per_user, buffer_id_offset)

        audio_buses_per_user = int(self.options.num_private_buses / self.max_logins)
        audio_bus_id_offset = (
            self.client_id * audio_buses_per_user + self.options.first_private_bus
        )
        self.audio_bus_ids = BlockAllocator(audio_buses_per_user, audio_bus_id_offset)

        control_buses_per_user = int(self.options.num_control_buses / self.max_logins)
        control_bus_id_offset = self.client_id * control_buses_per_user
        self.control_bus_ids = BlockAllocator(
            control_buses_per_user, control_bus_id_offset
        )

        # init I/O Buses
        self._output_bus = Bus(
            rate=BusRate.AUDIO,
            num_channels=self.options.num_output_buses,
            index=0,
            server=self,
        )
        self._input_bus = Bus(
            rate=BusRate.AUDIO,
            num_channels=self.options.num_input_buses,
            index=self.options.num_output_buses,
            server=self,
        )

        # load synthdefs of sc3nb
        self.load_synthdefs()

        # create default groups
        self.send_default_groups()
        self.sync()
        self._server_running = True

        for address in self.node_watcher.notification_addresses:
            self._osc_server.dispatcher.map(
                address, self.node_watcher.handle_notification
            )

        # note: init hooks should be last init step
        self.execute_init_hooks()
        self.sync()

        if with_blip:
            self.blip()
        print("Done.")

    def execute_init_hooks(self) -> None:
        """Run all init hook functions.

        This is automatically done when running free_all, init or connect_sclang.

        Hooks can be added using add_init_hook
        """
        _LOGGER.debug("Executing init hooks %s", self._server_init_hooks)
        for hook, args, kwargs in self._server_init_hooks:
            if args and kwargs:
                hook(*args, **kwargs)
            elif args:
                hook(*args)
            elif kwargs:
                hook(**kwargs)
            else:
                hook()

    def connect_sclang(self, port: int) -> None:
        """Connect sclang to the server

        This will add the "sclang" receiver and execute the init hooks

        Parameters
        ----------
        port : int
            Port of sclang (NetAddr.langPort)
        """
        self.add_receiver(name="sclang", ip_address="127.0.0.1", port=port)
        self.execute_init_hooks()

    def add_init_hook(
        self, hook: Callable[..., None], *args: Any, **kwargs: Any
    ) -> None:
        """Add a function to be executed when the server is initialized

        Parameters
        ----------
        hook : Callable[..., None]
            Function to be executed
        args : Any, optional
            Arguments given to function
        kwargs : Any, optional
            Keyword arguments given to function
        """
        self._server_init_hooks.append((hook, args, kwargs))

    def bundler(self, timetag=0, msg=None, msg_params=None, send_on_exit=True):
        """Generate a Bundler with added server latency.

        This allows the user to easly add messages/bundles and send it.

        Parameters
        ----------
        timetag : float
            Time at which bundle content should be executed.
            This servers latency will be added upon this.
            If timetag <= 1e6 it is added to time.time().
        msg_addr : str
            SuperCollider address.
        msg_params : list, optional
            List of parameters to add to message.
             (Default value = None)

        Returns
        -------
        Bundler
            bundler for OSC bundling.
        """
        return super().bundler(
            timetag=timetag + self.latency,
            msg=msg,
            msg_params=msg_params,
            send_on_exit=send_on_exit,
        )

    def blip(self) -> None:
        """Make a blip sound"""
        if self._volume.muted:
            warnings.warn("SCServer is muted. Blip will also be silent!")
        with self.bundler(0.15) as bundler:
            bundler.add(0.0, "/error", [0])
            bundler.add(
                0.0, "/s_new", ["s1", -1, 0, 0, "freq", 500, "dur", 0.1, "num", 1]
            )
            bundler.add(
                0.2, "/s_new", ["s2", -1, 0, 0, "freq", 1000, "amp", 0.05, "num", 2]
            )
            bundler.add(0.3, "/n_free", [-1])
            bundler.add(0.3, "/error", [1])

    def remote(self, address: str, port: int, with_blip: bool = True) -> None:
        """Connect to remote Server

        Parameters
        ----------
        address : str
            address of remote server
        port : int
            port of remote server
        with_blip : bool, optional
            make a sound when initialized, by default True
        """
        self._is_local = False
        self._scsynth_address = address
        self._scsynth_port = port
        self.init(with_blip=with_blip)
        self._has_booted = True

    def reboot(self) -> None:
        """Reboot this server

        Raises
        ------
        RuntimeError
            If this server is remote and can't be restarted.
        """
        if not self.is_local:
            raise RuntimeError("Can't reboot a remote Server")
        receivers = self._receivers  # save known receivers and restore them after boot
        self.quit()
        self.boot()
        receivers.update(
            self._receivers
        )  # update old receivers with possible new values
        self._receivers = receivers
        self.execute_init_hooks()
        self.sync()

    def ping(self):
        """Ping the server."""
        raise NotImplementedError

    # messages
    def quit(self) -> None:
        """Quits and tries to kill the server."""
        print("Quitting SCServer... ", end="")
        try:
            self.msg(MasterControlCommand.QUIT, bundle=False)
        except OSCCommunicationError:
            pass  # sending failed. scscynth maybe dead already.
        finally:
            super().quit()
            self._server_running = False
            if self._is_local:
                self._has_booted = False
                self.process.kill()
            print("Done.")

    def sync(self, timeout=5) -> bool:
        """Sync the server with the /sync command.

        Parameters
        ----------
        timeout : int, optional
            Time in seconds that will be waited for sync.
             (Default value = 5)

        Returns
        -------
        bool
            True if sync worked.
        """
        sync_id = randint(1000, 9999)
        msg = OSCMessage(MasterControlCommand.SYNC, sync_id)
        return sync_id == self.send(msg, timeout=timeout, bundle=False)

    def send_synthdef(self, synthdef_bytes: bytes):
        """Send a SynthDef as bytes.

        Parameters
        ----------
        synthdef_bytes : bytes
            SynthDef bytes
        wait : bool
            If True wait for server reply.
        """
        SynthDef.send(synthdef_bytes=synthdef_bytes, server=self)

    def load_synthdef(self, synthdef_path: str):
        """Load SynthDef file at path.

        Parameters
        ----------
        synthdef_path : str
            Path with the SynthDefs
        bundle : bool
            Wether the OSC Messages can be bundle or not.
            If True sc3nb will not wait for the server response, by default False
        """
        SynthDef.load(synthdef_path=synthdef_path, server=self)

    def load_synthdefs(
        self,
        synthdef_dir: Optional[str] = None,
        completion_msg: Optional[bytes] = None,
    ) -> None:
        """Load all SynthDefs from directory.

        Parameters
        ----------
        synthdef_dir : str, optional
            directory with SynthDefs, by default sc3nb default SynthDefs
        completion_msg : bytes, optional
            Message to be executed by the server when loaded, by default None
        """
        SynthDef.load_dir(
            synthdef_dir=synthdef_dir,
            completion_msg=completion_msg,
            server=self,
        )

    def notify(
        self,
        receive_notifications: bool = True,
        client_id: Optional[int] = None,
        timeout: float = 1,
    ) -> None:
        """Notify the server about this client.

        This provides the client id and max logins info needed for default groups.

        Parameters
        ----------
        receive_notifications : bool, optional
            Flag for receiving node notification from server, by default True
        client_id : int, optional
            Propose a client id, by default None
        timeout : float, optional
            Timeout for server reply, by default 1.0

        Raises
        ------
        RuntimeError
            If server has too many users.
        OSCCommunicationError
            If OSC communication fails.
        """
        flag = 1 if receive_notifications else 0
        client_id = client_id if client_id is not None else self._client_id
        msg = OSCMessage(
            MasterControlCommand.NOTIFY, [flag, client_id]
        )  # flag, clientID
        try:
            return_val = self.send(msg, timeout=timeout, bundle=False)
        except OSCCommunicationError as error:
            errors = self._get_errors_for_address(msg.address)
            if len(errors) > 0:
                last_error_value = errors[-1]
                if isinstance(last_error_value, tuple):
                    message, *rest = last_error_value
                else:
                    message = last_error_value
                    rest = None

                if "already registered" in message:
                    self._client_id = rest[0]
                    return  # only send client_id but not max logins
                elif "too many users" in message:
                    raise RuntimeError(
                        "scsynth has too many users. Can't notify."
                    ) from error
                elif "not registered" in message:
                    return  # ignore when we are already not notified anymore.
            raise error
        else:
            if receive_notifications:
                self._client_id, self._max_logins = return_val

    def free_all(self, root: bool = True) -> None:
        """Free all node ids.

        Parameters
        ----------
        root : bool, optional
            If False free only the default group of this client, by default True
        """
        group = self._root_node if root else self.default_group
        group.free_all()
        self.clear_schedule()
        if root:
            self.send_default_groups()
        else:
            self.default_group.new()
        self.execute_init_hooks()
        self.sync()

    def clear_schedule(self):
        """Send /clearSched to the server.

        This clears all scheduled bundles and removes all bundles from the scheduling queue.
        """
        self.msg(MasterControlCommand.CLEAR_SCHED)

    def send_default_groups(self) -> None:
        """Send the default groups for all clients."""
        client_ids = range(self._max_logins)

        def create_default_group(client_id) -> Group:
            return Group(
                nodeid=2 ** 26 * client_id + 1, target=0, server=self, new=True
            )

        self._default_groups = {
            client: create_default_group(client) for client in client_ids
        }

    @property
    def client_id(self):
        """The client id for this server"""
        return self._client_id

    @property
    def max_logins(self):
        """Maximum number of possible logins at server"""
        return self._max_logins

    @property
    def default_group(self):
        """This clients default group"""
        return self._default_groups[self._client_id]

    @property
    def input_bus(self) -> Bus:
        """This servers input Bus"""
        return self._input_bus

    @property
    def output_bus(self) -> Bus:
        """This servers output Bus"""
        return self._output_bus

    @property
    def volume(self) -> float:
        """Volume in dB."""
        return self._volume.volume

    @volume.setter
    def volume(self, volume):
        self._volume.volume = volume

    @property
    def muted(self) -> bool:
        """True if audio is muted"""
        return self._volume.muted

    @muted.setter
    def muted(self, muted):
        self._volume.muted = muted

    def mute(self) -> None:
        """Mute audio"""
        self._volume.mute()

    def unmute(self) -> None:
        """Set volume back to volume prior to muting"""
        self._volume.unmute()

    # Information and debugging
    def version(self) -> ServerVersion:
        """Server version information"""
        msg = OSCMessage(MasterControlCommand.VERSION)
        return ServerVersion._make(self.send(msg, bundle=False))

    def status(self) -> ServerStatus:
        """Server status information"""
        msg = OSCMessage(MasterControlCommand.STATUS)
        return ServerStatus._make(self.send(msg, bundle=False)[1:])

    def dump_osc(self, level: int = 1) -> None:
        """Enable dumping incoming OSC messages at the server process

        Parameters
        ----------
        level : int, optional
            Verbosity code, by default 1
            0   turn dumping OFF.
            1   print the parsed contents of the message.
            2   print the contents in hexadecimal.
            3   print both the parsed and hexadecimal representations.
        """
        msg = OSCMessage(MasterControlCommand.DUMP_OSC, [level])
        self.send(msg)

    def dump_tree(self, controls: bool = True, return_tree=False) -> Optional[str]:
        """Server process prints out current nodes

        Parameters
        ----------
        controls : bool, optional
            If True include control values, by default True
        return_tree : bool, optional
            If True return output as string, by default False

        Returns
        -------
        str
            If return_tree this is the node tree string.
        """
        if not self.is_local or self.process is None:
            warnings.warn("Server is not local or not booted. Use query_tree")
            return
        self.process.read()
        msg = OSCMessage(GroupCommand.DUMP_TREE, [0, 1 if controls else 0])
        self.send(msg, bundle=False)
        node_tree = self.process.read(expect="NODE TREE")
        print(node_tree)
        if return_tree:
            return node_tree

    def query_tree(self, include_controls: bool = True) -> Group:
        """Query all nodes at the server and return a NodeTree

        Parameters
        ----------
        include_controls : bool, optional
            If True include control values, by default True

        Returns
        -------
        NodeTree
            object containing all the nodes.
        """
        return self._root_node.query_tree(include_controls=include_controls)

    @property
    def peak_cpu(self) -> float:
        """Peak cpu usage of server process"""
        return self.status().peak_cpu

    @property
    def avg_cpu(self) -> float:
        """Average cpu usage of server process"""
        return self.status().peak_cpu

    @property
    def nominal_sr(self) -> float:
        """Nominal sample rate of server process"""
        return self.status().nominal_sr

    @property
    def actual_sr(self) -> float:
        """Actual sample rate of server process"""
        return self.status().actual_sr

    @property
    def num_synths(self) -> int:
        """Number of Synths in server tree"""
        return self.status().num_synths

    @property
    def num_groups(self) -> int:
        """Number of Groups in server tree"""
        return self.status().num_groups

    @property
    def num_ugens(self) -> int:
        """Number of UGens in server tree"""
        return self.status().num_ugens

    @property
    def num_synthdefs(self) -> int:
        """Number of SynthDefs known by server"""
        return self.status().num_synthdefs

    @property
    def addr(self) -> Tuple[str, int]:
        """Address (ip, port) of server"""
        return (self._scsynth_address, self._scsynth_port)

    @property
    def has_booted(self) -> bool:
        """If the server is booted"""
        return self._has_booted

    @property
    def is_running(self) -> bool:
        """If the server is running"""
        return self._server_running

    @property
    def unresponsive(self) -> bool:
        """If the server process is unresponsive"""
        try:
            self.status()
        except OSCCommunicationError:
            return True
        else:
            return False

    @property
    def is_local(self) -> bool:
        """If the server process is local"""
        return self._is_local

    @property
    def pid(self):
        """The process id of the local server process"""
        if self.is_local and self.process is not None:
            return self.process.popen.pid
        else:
            warnings.warn("Server is not local or not booted.")

    def _init_osc_communication(self):
        super().__init__(
            server_ip=LOCALHOST,
            server_port=SC3NB_DEFAULT_PORT,
            default_receiver_ip=LOCALHOST,
            default_receiver_port=self.options.udp_port,
        )
        # init msg queues
        self.add_msg_pairs(CMD_PAIRS)

        # /return messages from sclang callback
        self.returns = MessageQueue(ReplyAddress.RETURN_ADDR, preprocess_return)
        self.add_msg_queue(self.returns)

        # /done messages must be seperated
        self.dones = MessageQueueCollection(
            address=ReplyAddress.DONE_ADDR, sub_addrs=ASYNC_CMDS
        )
        self.add_msg_queue_collection(self.dones)

        self.fails = MessageQueueCollection(address=ReplyAddress.FAIL_ADDR)
        self.add_msg_queue_collection(self.fails)

        # set logging handlers
        self._osc_server.dispatcher.map(
            ReplyAddress.FAIL_ADDR, self._warn_fail, needs_reply_address=True
        )
        self._osc_server.dispatcher.map(
            ReplyAddress.WILDCARD_ADDR, self._log_message, needs_reply_address=True
        )

    def _get_errors_for_address(self, address: str):
        error_values = []
        if address in self.fails:
            while True:
                try:
                    error_values.append(self.fails.msg_queues[address].get(timeout=0))
                except Empty:
                    break
        return error_values

    def _log_repr(self):
        pid = f" pid={self.pid}" if self.is_local else ""
        return f"SCServer{self.addr}{pid}"

    def _log_message(self, sender, *params):
        params = str(params)
        if len(params) > 55:
            params = params[:55] + ".."
        _LOGGER.info(
            "%s got OSC msg from %s: %s",
            self._log_repr(),
            self._check_sender(sender),
            params,
        )

    def _warn_fail(self, sender, *params):
        _LOGGER.warning(
            "Error at %s from %s: %s",
            self._log_repr(),
            self._check_sender(sender),
            params,
        )

    def __repr__(self) -> str:
        if self.has_booted:
            status = f"addr={self.addr}, process={self.process}"
        else:
            status = "(not booted)"
        return f"<SCServer {status}>"



"""Module for handling a SuperCollider language (sclang) process."""
import logging
import re
import sys
import time
import warnings
from pathlib import Path
from queue import Empty
from typing import Any, NamedTuple, Optional, Sequence, Tuple

if sys.version_info < (3, 9):
    # `importlib.resources` backported to PY<37 as `importlib_resources`.
    import importlib_resources as libresources
else:
    # only PY>=39 `importlib.resources` offers .files.
    import importlib.resources as libresources

import sc3nb.resources
from sc3nb.process_handling import ALLOWED_PARENTS, Process, ProcessTimeout
from sc3nb.sc_objects.server import ReplyAddress, SCServer
from sc3nb.util import parse_pyvars, remove_comments, replace_vars

_LOGGER = logging.getLogger(__name__)


SCLANG_DEFAULT_PORT = 57120
SC3NB_SCLANG_CLIENT_ID = 0


class SynthArgument(NamedTuple):
    """Synth argument, rate and default value"""

    name: str
    rate: str
    default: Any


class SCLangError(Exception):
    """Exception for Errors related to SuperColliders sclang."""

    def __init__(self, message, sclang_output=None):
        super().__init__(message)
        self.sclang_output = sclang_output


class SCLang:
    """Class to control the SuperCollider Language Interpreter (sclang)."""

    def __init__(self) -> None:
        """Creates a python representation of sclang.

        Raises
        ------
        NotImplementedError
            When an unsupported OS was found.
        """
        self.process: Process = None
        self._server: Optional[SCServer] = None
        self.started: bool = False
        self._port: int = SCLANG_DEFAULT_PORT
        self._repl_return = "->"
        if sys.platform.startswith("linux") or sys.platform == "darwin":
            self.ending = "\n"
        elif sys.platform == "win32":
            self.ending = "\n\f"
        else:
            raise NotImplementedError("Unsupported OS {}".format(sys.platform))

    def start(
        self,
        sclang_path: Optional[str] = None,
        console_logging: bool = True,
        allowed_parents: Sequence[str] = ALLOWED_PARENTS,
        timeout: float = 10,
    ) -> None:
        """Start and initilize the sclang process.

        This will also kill sclang processes that does not have allowed parents.

        Parameters
        ----------
        sclang_path : Optional[str], optional
            Path with the sclang executable, by default None
        console_logging : bool, optional
            If True log sclang output to console, by default True
        allowed_parents : Sequence[str], optional
            parents name of processes to keep, by default ALLOWED_PARENTS
        timeout : float, optional
            timeout in seconds for starting the executable, by default 10

        Raises
        ------
        SCLangError
            When starting or initilizing sclang failed.
        """
        if self.started:
            warnings.warn("sclang arlready started")
            return
        print("Starting sclang process... ", end="")
        self.process = Process(
            executable="sclang",
            executable_path=sclang_path,
            console_logging=console_logging,
            allowed_parents=allowed_parents,
        )
        try:
            self.read(expect="Welcome to SuperCollider", timeout=timeout)
        except ProcessTimeout as process_timeout:
            if (
                process_timeout.output
                and "Primitive '_GetLangPort' failed" in process_timeout.output
            ):
                raise SCLangError(
                    "sclang could not bind udp socket. Try killing old sclang processes.",
                    process_timeout.output,
                ) from process_timeout

            raise process_timeout
        else:
            self.started = True
            print("Done.")
            self.init()

    def init(self):
        """Initialize sclang for sc3nb usage.

        This will register the /return callback in sclang and load the SynthDefs from sc3nb.

        This is done automatically by running start.
        """
        print("Registering OSC /return callback in sclang... ", end="")
        self.cmds(
            r"""
                "sc3nb - Registering OSC /return callback".postln;
                // NetAddr.useDoubles = true;
                r = r ? ();
                r.callback = { arg code, ip, port;
                    var result = code.interpret;
                    var addr = NetAddr.new(ip, port);
                    var prependSize = { arg elem;
                        if (elem.class == Array){
                            elem = [elem.size] ++ elem.collect(prependSize);
                        }{
                            elem;
                        };
                    };
                    var msgContent = prependSize.value(result);
                    addr.sendMsg(^replyAddress, msgContent);
                    result;  // result should be returned
                };""",
            pyvars={"replyAddress": ReplyAddress.RETURN_ADDR},
        )
        print("Done.")

        print("Loading default sc3nb SynthDefs... ", end="")
        self.load_synthdefs()
        print("Done.")

    def load_synthdefs(self, synthdefs_path: Optional[str] = None) -> None:
        """Load SynthDef files from path.

        Parameters
        ----------
        synthdefs_path : str, optional
            Path where the SynthDef files are located.
            If no path provided, load default sc3nb SynthDefs.
        """

        def _load_synthdef(path):
            self.cmds(
                r""" "sc3nb - Loading SynthDefs from ^synthdef_path".postln;
                PathName.new(^synthdefs_path).files.collect(
                { |path| (path.extension == "scsyndef").if({SynthDescLib.global.read(path); path;})}
                );""",
                pyvars={"synthdefs_path": path.as_posix()},
            )

        if synthdefs_path is None:
            ref = libresources.files(sc3nb.resources) / "synthdefs"
            with libresources.as_file(ref) as path:
                _load_synthdef(path)
        else:
            path = Path(synthdefs_path)
            if path.exists() and path.is_dir():
                _load_synthdef(path)
            else:
                raise ValueError(f"Provided path {path} does not exist or is not a dir")

    def kill(self) -> int:
        """Kill this sclang instance.

        Returns
        -------
        int
            returncode of the process.
        """
        print("Exiting sclang... ", end="")
        self.started = False
        try:
            self.cmds('"sc3nb - exiting sclang".postln; 0.exit;')
        except RuntimeError:
            pass
        else:
            self.process.popen.wait(1.0)  # let sclang exit
        print("Done.")
        return self.process.kill()

    def __del__(self):
        if self.started:
            self.kill()

    def __repr__(self) -> str:
        return f"<SCLang process={self.process}>"

    def cmd(
        self,
        code: str,
        pyvars: Optional[dict] = None,
        verbose: bool = True,
        discard_output: bool = True,
        get_result: bool = False,
        print_error: bool = True,
        get_output: bool = False,
        timeout: int = 1,
    ) -> Any:
        """Send code to sclang to execute it.

        This also allows to get the result of the code or the corresponding output.

        Parameters
        ----------
        code : str
            SuperCollider code to execute.
        pyvars : dict, optional
            Dictionary of name and value pairs of python
            variables that can be injected via ^name, by default None
        verbose : bool, optional
            If True print output, by default True
        discard_output : bool, optional
            If True clear output buffer before passing command, by default True
        get_result : bool, optional
            If True receive and return the
            evaluation result from sclang, by default False
        print_error : bool, optional
            If this and get_result is True and code execution fails
            the output from sclang will be printed.
        get_output : bool, optional
            If True return output. Does not override get_result
            If verbose this will be True, by default False
        timeout : int, optional
            Timeout in seconds for code execution return result, by default 1

        Returns
        -------
        Any
            if get_result=True,
                Result from SuperCollider code,
                not all SC types supported.
                When type is not understood this
                will return the datagram from the
                OSC packet.
            if get_output or verbose
                Output from SuperCollider code.
            if get_output and get_result=True
                (result, output)
            else
                None

        Raises
        ------
        RuntimeError
            If get_result is True but no OSCCommunication instance is set.
        SCLangError
            When an error with sclang occurs.
        """
        if pyvars is None:
            pyvars = parse_pyvars(code)
        code = replace_vars(code, pyvars)

        # cleanup command string
        code = remove_comments(code)
        code = re.sub(r"\s+", " ", code).strip()

        if get_result:
            if self._server is None:
                raise RuntimeError(
                    "get_result is only possible when connected to a SCServer"
                )
            # escape " and \ in our SuperCollider string literal
            inner_code_escapes = str.maketrans({ord("\\"): r"\\", ord('"'): r"\""})
            inner_code = code.translate(inner_code_escapes)
            # wrap the command string with our callback function
            code = r"""r['callback'].value("{0}", "{1}", {2});""".format(
                inner_code, *self._server.osc_server.server_address
            )

        if discard_output:
            self.empty()  # clean all past outputs

        # write command to sclang pipe \f
        if code and code[-1] != ";":
            code += ";"
        self.process.write(code + self.ending)

        return_val = None
        if get_result:
            try:
                return_val = self._server.returns.get(timeout)
            except Empty as empty_exception:
                out = self.read()
                if print_error:
                    print("ERROR: unable to receive /return message from sclang")
                    print("sclang output: (also see console) \n")
                    print(out)
                raise SCLangError(
                    "unable to receive /return message from sclang", sclang_output=out
                ) from empty_exception
        out = self.read(expect=self._repl_return, timeout=timeout)
        if verbose:
            if sys.platform == "darwin":
                out = out[out.find(";\n") + 2 :]  # skip code echo
            print(out)
        if get_output:
            return_val = (return_val, out) if get_result else out
        return return_val

    def cmdv(self, code: str, **kwargs) -> Any:
        """cmd with verbose=True"""
        if kwargs.get("pyvars", None) is None:
            kwargs["pyvars"] = parse_pyvars(code)
        return self.cmd(code, verbose=True, **kwargs)

    def cmds(self, code: str, **kwargs) -> Any:
        """cmd with verbose=False, i.e. silent"""
        if kwargs.get("pyvars", None) is None:
            kwargs["pyvars"] = parse_pyvars(code)
        return self.cmd(code, verbose=False, **kwargs)

    def cmdg(self, code: str, **kwargs) -> Any:
        """cmd with get_result=True"""
        if kwargs.get("pyvars", None) is None:
            kwargs["pyvars"] = parse_pyvars(code)
        return self.cmd(code, get_result=True, **kwargs)

    def read(
        self, expect: Optional[str] = None, timeout: float = 1, print_error: bool = True
    ) -> str:
        """Reads SuperCollider output from the process output queue.

        Parameters
        ----------
        expect : Optional[str], optional
            Try to read this expected string, by default None
        timeout : float, optional
            How long we try to read the expected string in seconds, by default 1
        print_error : bool, optional
            If True this will print a message when timed out, by default True

        Returns
        -------
        str
            output from sclang process.

        Raises
        ------
        timeout
            If expected output string could not be read before timeout.
        """
        try:
            return self.process.read(expect=expect, timeout=timeout)
        except ProcessTimeout as timeout_error:
            if print_error:
                error_str = "Timeout while reading sclang"
                if expect:
                    error_str += (
                        f'\nexpected: "{expect}"' " (sclang return value)"
                        if expect is self._repl_return
                        else ""
                    )
                error_str += "\noutput until timeout below: (also see console)"
                error_str += "\n----------------------------------------------\n"
                print(error_str + timeout_error.output)
            raise timeout_error

    def empty(self) -> None:
        """Empties sc output queue."""
        self.process.empty()

    def get_synth_description(self, synth_def):
        """Get a SynthDesc like description via sclang's global SynthDescLib.

        Parameters
        ----------
        synth_def : str
            SynthDef name

        Returns
        -------
        dict
            {argument_name: SynthArgument(rate, default)}

        Raises
        ------
        ValueError
            When synthDesc of synthDef can not be found.
        """
        code = r""" "sc3nb - Get SynthDesc of {{synthDef}}".postln;
                SynthDescLib.global['{{synthDef}}'].notNil.if({
                    SynthDescLib.global['{{synthDef}}'].controls.collect(
                        { | control | [control.name, control.rate, control.defaultValue] }
                    )
                })""".replace(
            "{{synthDef}}", synth_def
        )
        synth_desc = self.cmds(code, get_result=True, print_error=False)
        if synth_desc:
            return {
                s[0]: SynthArgument(s[0], *s[1:]) for s in synth_desc if s[0] != "?"
            }
        else:
            return None

    @property
    def addr(self) -> Tuple[str, int]:
        """The address of this sclang"""
        return ("127.0.0.1", self._port)

    @property
    def server(self) -> Optional[SCServer]:
        """The SuperCollider server connected to this sclang instance."""
        return self._server

    def connect_to_server(self, server: Optional[SCServer] = None):
        """Connect this sclang instance to the SuperCollider server.

        This will set Server.default and s to the provided remote server.

        Parameters
        ----------
        server : SCServer, optional
            SuperCollider server to connect. If None try to reconnect.

        Raises
        ------
        ValueError
            If something different from an SCServer or None was provided
        SCLangError
            If sclang failed to register to the server.
        """
        if server is None:
            server = self._server
        if not isinstance(server, SCServer):
            raise ValueError(f"Server must be instance of SCServer, got {type(server)}")
        code = r""" "sc3nb - Connecting sclang to scsynth".postln;
        Server.default=s=Server.remote('sc3nb_remote', NetAddr("{0}",{1}), clientID:{2});"""
        self.cmds(code.format(*server.addr, SC3NB_SCLANG_CLIENT_ID))
        try:  # if there are 'too many users' we failed. So the Exception is the successful case!
            self.read(expect="too many users", timeout=0.3, print_error=False)
        except ProcessTimeout:
            self._server = server
            self._port = self.cmdg("NetAddr.langPort", verbose=False)
            _LOGGER.info("Connecting %s with %s", self._server, self)
            self._server.connect_sclang(port=self._port)
            self._server.add_init_hook(self.cmds, "s.initTree;")
        else:
            raise SCLangError(
                "failed to register to the server (too many users)\n"
                "Restart the scsynth server with maxLogins >= 3 "
                "or specify a different server"
            )



import logging
import time

from sc3nb import Synth
from sc3nb.osc.osc_communication import Bundler, convert_to_sc3nb_osc
from tests.conftest import SCBaseTest


class OSCTest(SCBaseTest):

    __test__ = True

    def test_bundler_message_skip(self):
        value = 1337
        with Bundler() as bundler:
            bundler.add(0, "/sync", value)
        t0 = time.time()
        while self.sc.server.msg_queues["/synced"].size <= 0:
            self.assertLess(time.time() - t0, 0.2)
        with self.assertLogs(level=logging.WARNING) as log:
            sync_val = OSCTest.sc.server.sync()
        self.assertTrue("skipped value 1337" in log.output[-1])
        self.assertTrue(sync_val)

    def test_osc_conversion(self):
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's.' is unknown", msg="SynthDesc seems to be known"
        ):
            with Bundler(send_on_exit=False) as bundler:
                for num in range(6):
                    for i in range(num):
                        Synth("s1", dict(freq=1000 * num + i, dur=1 / 3))
                    bundler.wait(1)
        t0 = time.time() + 5
        original = list(bundler.messages(t0).items())
        converted = list(
            convert_to_sc3nb_osc(bundler.to_pythonosc(t0)).messages().items()
        )
        assert len(original) == len(converted)
        for i in range(len(original)):
            self.assertAlmostEqual(original[i][0], converted[i][0], places=6)
            assert len(original[i][1]) == len(converted[i][1])
            for j in range(len(original[i][1])):
                assert original[i][1][j].address == converted[i][1][j].address
                assert original[i][1][j].parameters == converted[i][1][j].parameters
